package org.openmuc.core.obis;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */


/**
 * 
 * A common representation for OBIS code strings according to IEC 62056-61.
 * 
 * Contains methods to - convert OBIS code strings to a unique representation -
 * test unique representations for equality - convert unique representation into
 * a complete OBIS code string - get descriptive strings for OBIS codes
 * 
 * @author mzillgit
 * 
 */

public class OBISCode {
	protected short medium; /* A */
	protected short channel; /* B */
	protected short measurement; /* C */
	protected short measurementMode; /* D */
	protected short payLevel; /* E */
	protected short previousValue; /* F */

	public OBISCode(short medium, short channel, short measurement, short measurementMode, short payLevel,
			short previousValue) {
		this.medium = medium;

		if ((channel == 1) || (channel == 0))
			this.channel = -1;
		else
			this.channel = channel;
		this.measurement = measurement;
		this.measurementMode = measurementMode;
		this.payLevel = payLevel;
		this.previousValue = previousValue;
	}

	@Deprecated
	public OBISCode() {
		medium = -1;
		channel = -1;
		measurement = -1;
		measurementMode = -1;
		payLevel = -1;
		previousValue = -1;
	}

	private boolean isChannelEqual(OBISCode code) {
		if (code.channel == this.channel)
			return true;

		if ((code.channel > -2 && code.channel < 2) && (this.channel > -2 && this.channel < 2))
			return true;

		return false;
	}

	private boolean isPreviousValueEqual(OBISCode code) {
		if (code.previousValue == this.previousValue)
			return true;

		if ((code.previousValue == 255 || code.previousValue == -1)
				&& (this.previousValue == 255 || this.previousValue == -1))
			return true;

		return false;
	}

	@Override
	/**
	 * 
	 * Compares to OBIS code representations
	 * 
	 * @return true if 
	 * 
	 */
	public boolean equals(Object obj) {
		OBISCode obisobj;

		if (obj.getClass() != this.getClass()) {
			System.out.println(obj.getClass().getName() + " is different from " + this.getClass().getName());
			return false;
		}

		obisobj = (OBISCode) obj;

		if (obisobj.medium == this.medium) {
			if (isChannelEqual(obisobj)) {
				if (obisobj.measurement == this.measurement) {
					if (obisobj.measurementMode == this.measurementMode) {
						if (obisobj.payLevel == this.payLevel) {
							if (isPreviousValueEqual(obisobj)) {
								return true;
							}
						}
					}
				}
			}
		}

		return false;
	} /* equals() */

	@Deprecated
	public short getMedium() {
		return medium;
	}

	@Deprecated
	public short getChannel() {
		return channel;
	}

	@Deprecated
	public short getMeasurement() {
		return measurement;
	}

	@Deprecated
	public short getMeasurementMode() {
		return measurementMode;
	}

	@Deprecated
	public short getPayLevel() {
		return payLevel;
	}

	@Deprecated
	public short getPreviousValue() {
		return previousValue;
	}

	/**
	 * 
	 * Create new instance by a given OBIS code string
	 * 
	 * @param obisCode
	 */
	public OBISCode(String obisCode) {
		parseOBISCodeString(obisCode);
	}

	/**
	 * Convert object to a OBIS code string according to IEC 62056-61 with all
	 * code elements
	 * 
	 * @return String representation of OBIS code
	 * 
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();

		if (medium != -1) {
			sb.append(medium);
			sb.append('-');
		}

		if (channel != -1) {
			sb.append(channel);
			sb.append(':');
		}
		else {
			sb.append(1);
			sb.append(':');
		}

		sb.append(measurement);
		sb.append('.');
		sb.append(measurementMode);

		if (payLevel != -1) {
			sb.append('.');
			sb.append(payLevel);
		}

		if (previousValue != -1) {
			sb.append('*');
			sb.append(previousValue);
		}
		else {
			sb.append("*255");
		}

		return sb.toString();
	} /* toString() */

	public OBISCode(byte[] code) {
		medium = (short) (0x00ff & (int) code[0]);
		channel = (short) (0x00ff & (int) code[1]);
		measurement = (short) (0x00ff & (int) code[2]);
		measurementMode = (short) (0x00ff & (int) code[3]);
		payLevel = (short) (0x00ff & (int) code[4]);
		previousValue = (short) (0x00ff & (int) code[5]);
	}

	private void parseOBISCodeString(String obisCode) {
		int pos;
		boolean prevValFound = false;
		String parseString = obisCode;

		if ((pos = Util.containsChar(parseString, '-')) > 0) {
			medium = Short.parseShort(parseString.substring(0, pos - 1));
			parseString = parseString.substring(pos);
		}
		else
			medium = -1;

		if ((pos = Util.containsChar(parseString, ':')) > 0) {
			channel = Short.parseShort(parseString.substring(0, pos - 1));
			if ((channel == 0) || (channel == -1)) {
				channel = -1;
			}
			parseString = parseString.substring(pos);
		}
		else
			channel = -1;

		if ((pos = Util.containsChar(parseString, '.')) > 0) {
			measurement = Short.parseShort(parseString.substring(0, pos - 1));
			parseString = parseString.substring(pos);
		}
		else {
			System.out.println("ParseError");
		}

		if ((pos = Util.containsChar(parseString, '.')) > 0) {
			int pos2;
			measurementMode = Short.parseShort(parseString.substring(0, pos - 1));

			pos2 = Util.containsChar(parseString, '*');

			if (pos2 > 0) {
				payLevel = Short.parseShort(parseString.substring(pos, pos2 - 1));
				previousValue = Short.parseShort(parseString.substring(pos2));
				prevValFound = true;
			}
			else {
				payLevel = Short.parseShort(parseString.substring(pos));
			}
		}
		else {
			int pos2;
			payLevel = -1;

			if ((pos2 = Util.containsChar(parseString, '*')) > 0) {
				measurementMode = Short.parseShort(parseString.substring(0, pos2 - 1));
				previousValue = Short.parseShort(parseString.substring(pos2));
				prevValFound = true;

				parseString = "";
			}
			else {
				measurementMode = Short.parseShort(parseString.substring(0));
			}
		}

		if (prevValFound == false) {
			if ((pos = Util.containsChar(parseString, '*')) > 0) {
				previousValue = Short.parseShort(parseString.substring(0, pos));
				prevValFound = true;
			}
			else {
				previousValue = -1;
			}
		}
	} /* parseOBISCodeString() */

	/**
	 * Tests if this OBIS code is in the range of manufacturer specific codes.
	 * 
	 */
	public boolean isManufacturerSpecific() {

		if (channel > 127 && channel < 200)
			return true;
		if (measurement > 127 && measurement < 200)
			return true;
		if (measurement == 240)
			return true;
		if (measurementMode > 127 && measurementMode < 255)
			return true;
		if (payLevel > 127 && payLevel < 255)
			return true;
		if (previousValue > 127 && previousValue < 255)
			return true;

		if (medium == 0 && measurement == 96)
			return true;
		if (medium == 1 && measurement == 96)
			return true;

		return false;
	}

	public String getDescription(int defaultMedium) {

		String meas;

		if ((medium == 1) || (medium == -1 && defaultMedium == 1)) {
			switch (measurement) {
			case 0:
				// TODO be more specific
				meas = "General purpose object";
				break;
			case 1:
				meas = "Active power+";
				break;
			case 2:
				meas = "Active power-";
				break;
			case 3:
				meas = "Reactive power+";
				break;
			case 4:
				meas = "Reactive power-";
				break;
			case 5:
				meas = "Reactive power QI";
				break;
			case 6:
				meas = "Reavtive power QII";
				break;
			case 7:
				meas = "Reactive power QIII";
				break;
			case 8:
				return "Reactive power QIV";
			case 9:
				return "Apparant power+ (QI + QIV)";
			case 10:
				return "Apparant power- (QII + QIII)";
			case 11:
				return "Current";
			case 12:
				return "Voltage";
			case 13:
				return "Power factor";
			case 14:
				return "Supply frequency";
			case 15:
				return "Active power+-";
			case 17:
				return "Active power QI";
			case 18:
				return "Active power QII";
			case 19:
				return "Active power QIII";
			case 20:
				return "Active power";
			case 21:
				return "L1 Active power+";
			case 22:
				return "L1 Active power-";
			case 23:
				return "L1 Reactive power+";
			}
		}

		return "";
	}

}
