package org.openmuc.core.datamanager.impl;

/*
 * Copyright Fraunhofer ISE, 2010,2011
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import org.apache.log4j.Logger;
import org.openmuc.core.datamanager.config.ChannelConfiguration;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.Unit;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.util.Date;

public class ChannelConfigurationImpl implements ChannelConfiguration {
	private static Logger logger = Logger.getLogger(ChannelConfigurationImpl.class);
	
	private String label;  /** unique name for data (like mux label) */
	private Integer storingPeriodInSeconds = null;      
	private Date storingStartTime = null;
	private String description = null;
	private String objectLocator = "";
	private Float absoluteChangeLevel = null;
	private Float relativeChangeLevel = null;	
	private Integer storageType = null;  /** none, instantaneous, mean value, value changed */
	private Double scaler = null; /** scaler - value received by driver will be multiplied by this value */
	private Double offset = null; /** offset - value that will be added to raw_value * scaler */
	private Double lowerLimit = null;
	private Double upperLimit = null;
	private Unit unit = Unit.UNKNOWN;
	
	private ChannelIdentifier dataID;
	
	public ChannelConfigurationImpl(String label) {
		this.label = label;
	}
	
	public ChannelConfigurationImpl(int readingPeriod,
			Date readingStartTime, String objectLocator, int storageType,
			Float absoluteChangeLevel, Float relativeChangeLevel,
			String label, String description) 
	{
		this.storingPeriodInSeconds = readingPeriod;
		this.storingStartTime = readingStartTime;
		this.objectLocator = objectLocator;
		this.absoluteChangeLevel = absoluteChangeLevel;
		this.relativeChangeLevel = relativeChangeLevel;
		this.label = label;
		this.description = description;
		this.storageType = storageType;
		this.dataID = new ChannelIdentifier(this.label, this.objectLocator);
	}
	
	public double getOffset() {
		if (offset == null) return 0.;
		else return offset;
	}

	public void setOffset(double offset) {
		this.offset = offset;
	}

	public double getScaler() {
		if (this.scaler == null) 
			return 1.0; 
		else 
			return scaler;
	}

	public void setScaler(double scaler) {
		this.scaler = scaler;
	}

	public Float getAbsoluteChangeLevel() {
		return absoluteChangeLevel;
	}

	public void setAbsoluteChangeLevel(float absoluteChangeLevel) {
		this.absoluteChangeLevel = absoluteChangeLevel;
	}

	public Float getRelativeChangeLevel() {
		return relativeChangeLevel;
	}
	
	public void setRelativeChangeLevel(float relativeChangeLevel) {
		this.relativeChangeLevel = relativeChangeLevel;
	}
	
	public int getStorageType() {
		return storageType;
	}
	
	public int getStoringPeriod() {
		return storingPeriodInSeconds;
	}

	public Date getStoringStartTime() {
		return storingStartTime;
	}
	
	public ChannelIdentifier getDataID() {
		if (dataID == null) {
			if ((this.objectLocator != null) && (this.label != null)) {
				dataID = new ChannelIdentifier(label, objectLocator);
			}
		}
		
		return dataID;
	}
	
	public String getDescription() {
		return description;
	}

	protected ChannelConfigurationImpl (Node channel) {
		
		NodeList childNodes = channel.getChildNodes();
		for(int i=0;i<childNodes.getLength();i++) {
			Node child = childNodes.item(i);
			try {
				if(child.getNodeName().equals("#text") || child.getNodeName().equals("#comment"))
					continue;
				else if(child.getNodeName().equals("storingPeriod")) {
					String sPeriod = child.getTextContent();
					this.storingPeriodInSeconds = timeToSec(sPeriod);
				}
				else if(child.getNodeName().equals("storingStartTime")) {
				
					// read date or time in ISO 8601 standard and UTC

					this.storingStartTime = XMLUtils.parseDateTime(child.getTextContent());

				}	
				else if(child.getNodeName().equals("objectLocator"))
					this.objectLocator = child.getTextContent();
				else if(child.getNodeName().equals("label")) 
					this.label = child.getTextContent();
				else if(child.getNodeName().equals("description")) 
					description = child.getTextContent();
				else if(child.getNodeName().equals("storageType")) {
					String sType = child.getTextContent();
					if(sType.equals("none"))
						this.storageType = ChannelConfiguration.STORE_TYPE_NONE;
					else if(sType.equals("instantaneous"))
						this.storageType = ChannelConfiguration.STORE_TYPE_INST;
					else if(sType.equals("mean value") || (sType.equals("mean")))
						this.storageType = ChannelConfiguration.STORE_TYPE_MEAN;
					else if(sType.equals("value changed") || (sType.equals("changed")))
						this.storageType = ChannelConfiguration.STORE_TYPE_VALUE_CHANGED;
					else
						logger.warn("storageType "+sType+" not found! Use none, instantaneous, mean value or value changed");
						
				}
				else if (child.getNodeName().equals("unit")) {
					String unitString = child.getTextContent();
					boolean found = false;
					
					for (Unit unit : Unit.values()) {
						if (unit.name().equals(unitString.toUpperCase())) {
							this.unit = unit;
							found = true;
							break;
						}
					}
					
					if (!found)
						logger.warn("Unknown unit: " + unitString);
				}
				else if (child.getNodeName().equals("scaler")) 
					this.scaler = Double.valueOf(child.getTextContent());
				else if (child.getNodeName().equals("lowerLimit")) 
					this.lowerLimit = Double.valueOf(child.getTextContent());
				else if (child.getNodeName().equals("upperLimit")) 
					this.upperLimit = Double.valueOf(child.getTextContent());
				else if (child.getNodeName().equals("offset"))
					this.offset = Double.valueOf(child.getTextContent());
				else if(child.getNodeName().equals("absoluteChangeLevel"))
					this.absoluteChangeLevel = Float.valueOf(child.getTextContent());
				else if(child.getNodeName().equals("relativeChangeLevel"))
					this.relativeChangeLevel = Float.valueOf(child.getTextContent());
				else
					logger.warn(child.getNodeName()+" is no valid entry");
			}catch(NumberFormatException e) { 
				e.printStackTrace();
			} catch (DOMException e) {
				e.printStackTrace();
			} 
		}
		
		if ((this.objectLocator != null) && (this.label != null)) {
			this.dataID = new ChannelIdentifier(this.label, this.objectLocator);
		}
		
		if (this.storageType == null) this.storageType = ChannelConfiguration.STORE_TYPE_NONE;
	}

	public Double getLowerLimit() {
		return lowerLimit;
	}

	public void setLowerLimit(Double lowerLimit) {
		this.lowerLimit = lowerLimit;
	}

	public Double getUpperLimit() {
		return upperLimit;
	}

	public void setUpperLimit(Double upperLimit) {
		this.upperLimit = upperLimit;
	}

	private int timeToSec(String time) throws NumberFormatException
	{
		int sec;
		
		if (Character.isDigit(time.charAt(time.length() - 1))) {
			sec = Integer.parseInt(time);
		}
		else {
			sec = Integer.parseInt(time.substring(0,time.length()-1));
		}
		
		switch(time.charAt(time.length()-1)) {
		case 's':
			break;
		case 'm':
			sec *= 60;
			break;
		case 'h':
			sec *= 60*60;
			break;
		case 'd':
			sec *= 60*60*24;
			break;
		case 'w':
			sec *= 60*60*24*7;
			break;
		case 'M':
			sec *= 60*60*24*30;
			break;
		case 'y':
			sec *= 60*60*24*365;
			break;
		default:
			/* default is seconds */
		}
		return sec;
	}
	

	public String getLabel() {
		return label;
	}

	public void setLabel(String label) {
		this.label = label;
	}

	public String getObjectLocator() {
		return objectLocator;
	}

	public void setObjectLocator(String objectLocator) {
		this.objectLocator = objectLocator;
	}

	public void setStoringPeriod(Integer storingPeriod) {
		this.storingPeriodInSeconds = storingPeriod;
	}

	public void setStoringStartTime(Date storingStartTime) {
		this.storingStartTime = storingStartTime;
	}

	public void setDescription(String description) {
		this.description = description;
	}

	public void setRelativeChangeLevel(Float relativeChangeLevel) {
		this.relativeChangeLevel = relativeChangeLevel;
	}

	public void setStorageType(Integer storageType) {
		this.storageType = storageType;
	}

	public void setScaler(Double scaler) {
		this.scaler = scaler;
	}

	public void persist(Element parent, Document document) {
		Element channel = getChannelWithLabel(parent, this.label);
		
		if (channel == null) {
			channel = document.createElement("channel");
			parent.appendChild(channel);
		}
		
		XMLUtils.updateTagWithValue(channel, document, "label", this.label);
		
		if (this.description != null)
			XMLUtils.updateTagWithValue(channel, document, "description",
					this.description);
		
		if (this.objectLocator != null)
			XMLUtils.updateTagWithValue(channel, document, "objectLocator", 
					this.objectLocator);
		
		if (this.storingPeriodInSeconds != null)
			XMLUtils.updateTagWithValue(channel, document, "storingPeriod", 
					this.storingPeriodInSeconds + "s");
		
		if (this.storageType != null)
			XMLUtils.updateTagWithValue(channel, document, "storageType", 
					XMLUtils.storageType(this.storageType));
		
		if (this.unit != null) {
			XMLUtils.updateTagWithValue(channel, document, "unit", this.unit.name());
		}
		
		if (this.storingStartTime != null)
			XMLUtils.updateTagWithValue(channel, document, "storingStartTime", 
					XMLUtils.dateToString(this.storingStartTime));
		
		if (this.scaler != null)
			XMLUtils.updateTagWithValue(channel, document, "scaler", this.scaler.toString());

		if (this.offset != null)
			XMLUtils.updateTagWithValue(channel, document, "offset", this.offset.toString());
		//TODO add all missing tags
	}
	
	private Element getChannelWithLabel(Element parent, String label) {
		NodeList nodeList = parent.getElementsByTagName("channel");
		
		// for compatibility with old configuration files:
		if (nodeList == null) nodeList = parent.getElementsByTagName("register");
		
		if (nodeList != null) {
			Node node = null;
			
			for (int i = 0; i < nodeList.getLength(); i++) {
				node = nodeList.item(i);
				if (hasLabelWithValue((Element) node, label)) return (Element) node;
			}
		}
		return null;
	}
	
	private boolean hasLabelWithValue(Element parent, String value) {
		NodeList nodeList = parent.getElementsByTagName("label");
		if (nodeList != null) {
			for (int i = 0; i < nodeList.getLength(); i++) {
				Node node = nodeList.item(i);
				if (node.getTextContent().equals(value)) return true;
			}
		}
		
		return false;
	}

	public void setStoringPeriod(int storingPeriod) {
		this.storingPeriodInSeconds = storingPeriod;
	}

	public void setStorageType(int storageType) {
		this.storageType = storageType;
	}

	public Unit getUnit() {
		return unit;
	}

	public void setUnit(Unit unit) {
		this.unit = unit;
	}
}
