package org.openmuc.core.datamanager.impl;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.log4j.Logger;
import org.openmuc.core.datamanager.config.ConfigurationListener;
import org.openmuc.core.datamanager.config.ConfigurationService;
import org.openmuc.core.datamanager.config.DeviceConfiguration;
import org.openmuc.core.datamanager.config.PlantConfiguration;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class Configuration implements ConfigurationService {

	List<WeakReference<ConfigurationListener>> listeners;
	
	protected Document doc;
	protected Element root;
	
	private static Logger logger = Logger.getLogger(Configuration.class);

	private File configFile;
	private boolean persistable = true; 
	
	private PlantConfigurationImpl plantConfiguration = null;
	
	private List<DeviceConfiguration> devices;
	
	private DataManagerImpl dataManager;

	public Configuration(String configFileName) {		
		if (configFileName == null || configFileName.trim().equals("")) {
			logger.info("Configuration file not defined! use \n-Dorg.openmuc.metering.configfile=<filename>\n to define it");
			logger.info("Starting with empty configuration.");
			logger.info("Configuration cannot be persisted!");
			persistable = false;
		}
		
		listeners = new LinkedList<WeakReference<ConfigurationListener>>();
		
		devices = new ArrayList<DeviceConfiguration>();
		
		configFile = new File(configFileName);
		
		if (configFile.exists()) {
			readConfigFile();
			parseConfigurationDocument();
		}
		else {
			createConfigFile();
		}
			
		for(DeviceConfiguration device: devices)
			logger.info("Found meter: " + device);
	}
	
	public Configuration(InputStream stream) {
		listeners = new LinkedList<WeakReference<ConfigurationListener>>();
		devices = new ArrayList<DeviceConfiguration>();
		
		doc = XMLUtils.getDocumentFromStream(stream);
		root = doc.getDocumentElement();
		parseConfigurationDocument();
	}
	
	protected void setDataManager(DataManagerImpl dataManager) {
		this.dataManager = dataManager;
	}
	
	public void persist() {
		if (plantConfiguration != null) {
			plantConfiguration.persist(root, doc);
		}
		
		for (DeviceConfiguration device : devices) {
			((DeviceConfigurationImpl) device).persist(root, doc);
		}

		writeConfigurationToFile();
		reloadConfiguration();
	}
	
	public DeviceConfiguration createDevice(String name) {
		DeviceConfigurationImpl newDevice = new DeviceConfigurationImpl(this, name);
		devices.add(newDevice);

		return newDevice;
	}
	
	private void createConfigFile() {
		DocumentBuilderFactory docBFac;
		DocumentBuilder docBuild;

		docBFac = DocumentBuilderFactory.newInstance();

		try {
			docBuild = docBFac.newDocumentBuilder();
			doc = docBuild.newDocument();
			root = doc.createElement("configuration");
			doc.appendChild(root);
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		} 
	}

	private void writeConfigurationToFile() {
		Transformer transformer;
		try {
			transformer = TransformerFactory.newInstance().newTransformer();
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");

			StreamResult result = new StreamResult(new FileWriter(configFile));
			DOMSource source = new DOMSource(doc);
			transformer.transform(source, result);
		} catch (TransformerConfigurationException e) {
			e.printStackTrace();
		} catch (TransformerFactoryConfigurationError e) {
			e.printStackTrace();
		} catch (TransformerException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void parseConfigurationDocument() {
		NodeList childlist = root.getChildNodes();
		
		for(int i=0;i<childlist.getLength();i++) {
			Node child = childlist.item(i);
			if(child.getNodeName().equals("#text") || child.getNodeName().equals("#comment"))
				continue;
			if (child.getNodeName().equals("meter") || child.getNodeName().equals("device")) {
				DeviceConfigurationImpl device = new DeviceConfigurationImpl(this, child);
				if (device != null)
					devices.add(device);
			}
			else if (child.getNodeName().equals("plant")) {
				this.plantConfiguration = new PlantConfigurationImpl(this, child);
			}
			else {
				logger.info(child.getNodeName() + " is no valid entry in config-file");
			}
		}
		
		sendConfigurationChangedEvent();
	}

	public void reloadConfiguration() {
		devices = new ArrayList<DeviceConfiguration>();
		readConfigFile();
		parseConfigurationDocument();
	}

	public List<DeviceConfiguration> getDevices() {
		return Collections.unmodifiableList(devices);
	}
		
	public DeviceConfiguration getDeviceByName(String name) {
		for (DeviceConfiguration deviceConfig : devices) {
			if (deviceConfig.getName().equals(name)) 
				return deviceConfig;
		}
		
		return null;
	}
	
	private void readConfigFile() {
		doc = XMLUtils.getDocumentFromFile(configFile);
		root = doc.getDocumentElement();
	}

	public void deleteDevice(String name) {
		DeviceConfiguration device = getDeviceByName(name);
		devices.remove(device);
		
		/* Remove meter from XML DOM */
		XMLUtils.deleteElementWithAttributeValue(doc, "meter", "name", name);
		XMLUtils.deleteElementWithAttributeValue(doc, "device", "name", name);
	}
	
	protected Document getXmlDocument() {
		return doc;
	}
	
	public void sendConfigurationChangedEvent() {
		if (dataManager != null) dataManager.restart();
		
		for (WeakReference<ConfigurationListener> listener : listeners) {
			ConfigurationListener configListener = listener.get();
			if (configListener != null)
				configListener.configurationChanged();
		}
	}

	public void registerListener(ConfigurationListener listener) {
		listeners.add(new WeakReference<ConfigurationListener>(listener));
	}

	public PlantConfiguration getPlantConfiguration() {
		return plantConfiguration;
	}
}
