package org.openmuc.core.datamanager.impl;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.apache.log4j.Logger;
import org.openmuc.core.datamanager.DatabaseInterface;
import org.openmuc.core.datamanager.config.ChannelConfiguration;
import org.openmuc.core.measurement.Value;

public class SQLiteDatabaseLayerProvider implements DatabaseInterface {

	private static Logger logger = Logger.getLogger(SQLiteDatabaseLayerProvider.class);
	
	private class RegisterIDCacheEntry {
		public RegisterIDCacheEntry(int id, String label) {
			this.id = id;
			this.label = label;
		}

		private int id;
		private String label;
	}

	private class ReadingInsert {
		public ReadingInsert(int meterId, long timestamp, double value) {
			this.meterId = meterId;
			this.timestamp = timestamp;
			this.value = value;
		}

		int meterId;
		long timestamp;
		double value;
	}

	private class InsertQueueWalker implements Runnable {
		private ConcurrentLinkedQueue<ReadingInsert> readingInserts;
		private boolean running;

		private PreparedStatement prepInsertReading;

		public InsertQueueWalker(ConcurrentLinkedQueue<ReadingInsert> readingInserts) {
			this.readingInserts = readingInserts;

			/* Prepare statements */
			try {
				prepInsertReading = con
						.prepareStatement("insert into MeterReading(idMeterRegister,time,reading,flag) values (?, ?, ?,'f');");
			} catch (SQLException e) {
				e.printStackTrace();
			}

		}

		private void insertBatchList(Vector<ReadingInsert> batchList) throws SQLException {
			for (ReadingInsert reading : batchList) {
				try {
					prepInsertReading.setInt(1, reading.meterId);
					prepInsertReading.setLong(2, reading.timestamp);
					prepInsertReading.setDouble(3, reading.value);
					prepInsertReading.addBatch();
				} catch (SQLException e) {
					logger.error("InsertQueueWalker - prepare insert " + e.getMessage());
				}
			}
		}

		public void run() {
			running = true;

			while (running) {

				ReadingInsert reading;
				Vector<ReadingInsert> batchList = null;

				try {Thread.sleep(50000);} catch (InterruptedException e) {}

				int openInsertsCounter = 0;

				long curTime = System.currentTimeMillis();

				while (!readingInserts.isEmpty()) {
					reading = readingInserts.poll();

					if (reading != null) {
						if (batchList == null)
							batchList = new Vector<ReadingInsert>();

						batchList.add(reading);
					} else
						System.out.println("ERROR: reading = null");

					openInsertsCounter++;
				}

				int trials = 0;

				while ((openInsertsCounter > 0) && (trials < 3)) {
					System.out.println("Storing " + openInsertsCounter + " values ...");
					synchronized (stat) {
						try {
							insertBatchList(batchList);
							con.setAutoCommit(false);
							prepInsertReading.executeBatch();
							con.setAutoCommit(true);
							logger.info("Stored " + openInsertsCounter + " values in "
									+ (System.currentTimeMillis() - curTime) + " ms");
							break;
						} catch (SQLException e) {
							if (e.toString().contains("database or disk is full")) {
								logger.info("DATABASE is full! Delete oldest rows...");
								cleanUp();
								logger.info("cleanup finished");
							} else {
								logger.error("store " + openInsertsCounter + " values: " + e.getMessage());
								e.printStackTrace();
							}
						}

						trials++;
					} /* synchronized (stat) */
				}
			}
		} /* run() */

	} /* class InsertQueueWalker */

	private ConcurrentLinkedQueue<ReadingInsert> readingInserts = new ConcurrentLinkedQueue<ReadingInsert>();

	private Vector<RegisterIDCacheEntry> regIdCache = new Vector<RegisterIDCacheEntry>();

	private Connection con;
	private Statement stat;

	private int maxFileSize = 1024; /* maximum file size in kilobytes */

	private String tempStoreDir = null; /*
										 * Directory where sqlite store its
										 * temporary files
										 */

	private static String filename = null; /*
											 * Full path and filename of
											 * database
											 */

	private String connectionStr;

	public static boolean setFilename(String name) {
		if (filename == null) {
			filename = name;
			return true;
		} else
			return false;
	}

	private InsertQueueWalker insertQueueWalker;

	public SQLiteDatabaseLayerProvider() {

		try {
			Class.forName("org.sqlite.JDBC");
		} catch (ClassNotFoundException e) {
			logger.error(e.getMessage());
		}

		if (filename == null) {
			filename = System.getProperty("org.openmuc.mux.dbfile");
		}

		if (filename == null) {
			logger.error("Define property org.openmuc.mux.dbfile!");
			filename = "mux.db";
		}

		tempStoreDir = System.getProperty("org.openmuc.mux.dbtempdir");

		if (tempStoreDir == null) {
			logger.error("Define property org.openmuc.mux.dbtempdir!");
			tempStoreDir = "/tmp";
		}

		String maxFileSize = System.getProperty("org.openmuc.mux.maxdbfilesize");

		if (maxFileSize != null) {
			logger.info("Set maximum file size to " + Integer.parseInt(maxFileSize));
			this.maxFileSize = Integer.parseInt(maxFileSize);
		}

		this.connectionStr = "jdbc:sqlite:" + filename;

		try {
			con = DriverManager.getConnection(connectionStr);
		} catch (SQLException e) {
			logger.error(e.getMessage());
		}

		try {
			stat = con.createStatement();
		} catch (SQLException e) {
			logger.error(e.getMessage());
		}

		/*
		 * Set maximum block size of database file. Any insert after this limit
		 * is reached will cause a SQLException "database or disk is full".
		 */
		setMaxPageCount(this.maxFileSize);

		/* Set necessary temp file options for sqlite */
		setTempFileOptions();

		/* Create tables if they don't exist */
		try {
			stat.executeUpdate("CREATE TABLE if not exists `MeterRegister` "
					+ "(`idMeterRegister` INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, "
					+ "`label` varchar(129) UNIQUE, `valueType` varchar(64), `description` varchar(100)  );");
		} catch (SQLException e) {
			logger.error(e.getMessage());
		}

		try {
			stat
					.executeUpdate("CREATE TABLE if not exists `MeterReading` (`idMeterReading` INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, `idMeterRegister` INTEGER, `time` INTEGER NOT NULL, `reading` REAL, `flag` varchar(1), FOREIGN KEY (`idMeterRegister`) REFERENCES MeterRegister('idMeterRegister') ON DELETE CASCADE );");
		} catch (SQLException e) {
			logger.error(e.getMessage());
		}

		/*
		 * Create index on idMeterRegister and time to improve select
		 * performance
		 */
		try {
			stat.executeUpdate("create index if not exists idandtimeindex on MeterReading ( idMeterRegister, time );");
		} catch (SQLException e) {
			logger.error(e.getMessage());
			e.printStackTrace();
		}

		insertQueueWalker = new InsertQueueWalker(readingInserts);

		Thread thread = new Thread(insertQueueWalker);
		thread.start();
	}

	private int getRowCount() {
		String queryStr = "select count() from MeterReading;";

		synchronized (stat) {
			ResultSet rs;
			try {
				rs = stat.executeQuery(queryStr);

				if (rs.next()) {
					return rs.getInt(1);
				}
			} catch (SQLException e) {
				logger.error(e.getMessage());
			}
		}

		return -1;
	}

	/* Remove about 30 % of database rows in table MeterReading */
	private void cleanUp() {
		int rows = getRowCount();

		int deleteRows = (int) (rows * 0.3);

		logger.debug("Delete " + deleteRows + " rows from MeterReading");

		String queryStr = "delete from MeterReading where idMeterReading in (select idMeterReading from MeterReading order by ROWID limit "
				+ deleteRows + ");";

		long task = System.currentTimeMillis();

		logger.info("Clean up database - remove old rows - task: " + task);
		simpleSQLQuery(queryStr);
		logger.info("Clean up database - remove completed --> starting vacuum - task: " + task);

		/* Compress database */
		queryStr = "vacuum";
		simpleSQLQuery(queryStr);		
		logger.info("Clean up database - vacuum completed - task: " + task);
	} /* cleanUp() */

	private void setMaxPageCount(int maxPageCount) {
		String queryStr = "pragma max_page_count=" + maxPageCount + ";";
		simpleSQLQuery(queryStr);
	}

	private void setTempFileOptions() {
		String queryStr = "pragma temp_store=1;";
		simpleSQLQuery(queryStr);

		queryStr = "pragma temp_store_directory=\"" + tempStoreDir + "\";";
		simpleSQLQuery(queryStr);
	} /* setTempFileOptions() */

	private void simpleSQLQuery(String queryStr) {
		synchronized (stat) {
			try {
				logger.debug("SQL: " + queryStr);
				stat.execute(queryStr);
			} catch (SQLException e) {
				logger.error("SQL: " + queryStr + " : " + e.getMessage());
			}
		}
	}

	public void addChannel(ChannelConfiguration channel) {
		String description = "";
		int regId;

		if (channel.getDescription() != null)
			description = channel.getDescription();

		String queryStr = "insert into MeterRegister(label,valueType,description) values (\"" + channel.getLabel()
				+ "\",\"" + "empty" + "\",\"" + description + "\");";

		/* Check if register with this label already exists */
		regId = getMeterRegisterID(channel.getLabel());

		synchronized (stat) {

			if (regId == -1) {
				/* Insert new register into database */
				try {
					stat.execute(queryStr);
				} catch (SQLException e) {
					logger.error("addRegister: " + e.getMessage() + " Error code: " + e.getErrorCode());
				}

			}
		}
	}

	private int getMeterRegisterID(String label) {

		for (RegisterIDCacheEntry entry : regIdCache) {
			if (entry.label == label)
				return entry.id;
		}

		int ret = -1;

		String queryStr = "select \"idMeterRegister\" from MeterRegister where label=\"" + label + "\";";

		synchronized (stat) {
			try {
				ResultSet rs;

				rs = stat.executeQuery(queryStr);

				if (rs.next()) {
					ret = rs.getInt("idMeterRegister");
					regIdCache.add(new RegisterIDCacheEntry(ret, label));
				}

				rs.close();
			} catch (SQLException e) {
				logger.error("getMeterRegisterID: " + e.getMessage() + " Error code: " + e.getErrorCode());
			}
		}

		return ret;
	}

	public void insertValue(String label, Value reading) throws IOException {
		int meterId = this.getMeterRegisterID(label);

		readingInserts.add(new ReadingInsert(meterId, reading.getTimestamp(), reading.getValue()));
	}

	public List<Value> getValues(String label, long startTime) {
		int meterId;

		meterId = getMeterRegisterID(label);

		if (meterId != -1) {
			List<Value> res = new Vector<Value>();

			double reading;
			long timestamp;
			ResultSet rs;

			String queryStr = "select * from MeterReading where idMeterRegister=\"" + meterId + "\" and time>=\""
					+ startTime + "\";";

			logger.debug("getValues(): " + queryStr);

			synchronized (stat) {

				try {

					stat = con.createStatement();

					rs = stat.executeQuery(queryStr);
					while (rs.next()) {
						reading = rs.getDouble("reading");
						timestamp = rs.getLong("time");
						res.add(new Value(reading, timestamp));
					}

					rs.close();
					stat.close();
				} catch (SQLException e) {
					logger.error(e.getMessage());
				}
			}

			return res;
		} else
			return null;
	} /* getValues(start) */

	/* (non-Javadoc)
	 * @see org.openmuc.core.datamanager.DatabaseLayerInterface#getValues(java.lang.String, long, long)
	 */
	public List<Value> getValues(String label, long startTime, long endTime) {
		int meterId;

		meterId = getMeterRegisterID(label);

		if (meterId != -1) {
			List<Value> res = new Vector<Value>();

			double reading;
			long timestamp;
			ResultSet rs;

			String queryStr = "select * from MeterReading where idMeterRegister=\"" + meterId + "\" and time>=\""
					+ startTime + "\" and time <= \"" + endTime + "\";";

			logger.debug("getValues(): " + queryStr);

			synchronized (stat) {

				try {

					stat = con.createStatement();

					rs = stat.executeQuery(queryStr);
					while (rs.next()) {
						reading = rs.getDouble("reading");
						timestamp = rs.getLong("time");
						res.add(new Value(reading, timestamp));
					}

					rs.close();
					stat.close();
				} catch (SQLException e) {
					logger.error(e.getMessage());
				}
			}

			return res;
		} else
			return null;
	} /* getValues(start, stop) */

	/* (non-Javadoc)
	 * @see org.openmuc.core.datamanager.DatabaseLayerInterface#getValue(java.lang.String, long)
	 */
	public Value getValue(String label, long timestamp) {
		int meterId;

		logger.debug("ENTER getValue(" + label + "," + timestamp + ")");

		meterId = getMeterRegisterID(label);

		logger.debug("meterID: " + meterId);

		if (meterId != -1) {
			double reading;
			ResultSet rs;

			String queryStr = "select reading from MeterReading where idMeterRegister=\"" + meterId + "\" and time=\""
					+ timestamp + "\";";

			logger.debug("getValue(): " + queryStr);

			synchronized (stat) {

				try {

					stat = con.createStatement();
					rs = stat.executeQuery(queryStr);
					if (rs.next()) {
						reading = rs.getDouble("reading");
						rs.close();
						stat.close();
						return new Value(reading, timestamp);
					}

					rs.close();
					stat.close();
				} catch (SQLException e) {
					logger.error(e.getMessage() + " - Code: " + e.getErrorCode());
				}
			}
		}
		
		return null;
	} /* getValue() */

}
