package org.openmuc.mux.dbprovicer.slotsdb.test;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Simon Fey
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;

import org.openmuc.core.measurement.State;
import org.openmuc.core.measurement.Value;
import org.openmuc.mux.dbprovider.slotsdb.SlotsDatabase;
import org.openmuc.mux.dbprovider.slotsdb.FileObjectList;
import org.openmuc.mux.dbprovider.slotsdb.FileObjectProxy;

import junit.framework.TestCase;

public class FileObjectProxyTest extends TestCase {

	private FileObjectProxy fop;
	
	protected void setUp() throws Exception {
		super.setUp();
		fop = new FileObjectProxy(SlotsDatabase.DB_TEST_ROOT_FOLDER);
		
		//delete all files in test folder!
		deleteFiles(new File(SlotsDatabase.DB_TEST_ROOT_FOLDER));
	}
	
	private void deleteFiles(File f) {
		  if (f.isDirectory()) {
		    for (File c : f.listFiles())
		    	deleteFiles(c);
		  }
		  f.delete();
	}
	
	public void testAppendValue1Day() throws IOException {
		long starttime = 1293836400000L; //1.1.2011 0.00 Uhr
		long stepIntervall = 3600000; //1-hour steps
		String label = "VoltageL1";
		String strDate = new SimpleDateFormat("yyyyMMdd").format(new Date(starttime));

		/*
		 * Appending 1 whole day - 1 Hour steps
		 */
		for(int i = 0; i<24; i++) {
			fop.appendValue(label, 1D/(i+1), starttime + (stepIntervall*i), (byte)i, stepIntervall);
		}
		
		File toCheck = new File(SlotsDatabase.DB_TEST_ROOT_FOLDER + strDate + "/" + label + "/" + starttime + SlotsDatabase.FILE_EXTENSION);
		DataInputStream dis = new DataInputStream(new FileInputStream(toCheck));
		
		assertEquals(starttime, dis.readLong());
		assertEquals(stepIntervall, dis.readLong());
		
		/*
		 * Comparing every value from whole day
		 */
		for(int i = 0; i< 24; i++) {
			assertEquals(1D/(i+1), dis.readDouble());
			assertEquals((byte)i,dis.readByte());
		}
		if(dis.available() != 0)
			fail("All 24 Values are read. File should end here!");
		
		/**
		 * Last Timestamp in Object should be from 1. Jan 23:00:00 o'clock (1293919200000)
		 */
		assertEquals(1293919200000L, new FileObjectList(SlotsDatabase.DB_TEST_ROOT_FOLDER + strDate + "/" + label).getCurrentFileObject().getTimestampForLatestValue());
	}
	
	public void testAppendValue1DayWithIntervallChange() throws Exception {
		long starttime = 1293836400000L; //1.1.2011 0.00 Uhr
		long stepIntervall60min = 3600000; //1-hour steps
		long stepIntervall30min = 1800000; //30-min steps
		long stepIntervall15min = 900000;  //15-min steps
		long stepIntervall60minagain = 3600000; //1-hour steps
		
		String label = "VoltageL1";
		String strDate = new SimpleDateFormat("yyyyMMdd").format(new Date(starttime));

		/*
		 * Appending 1 whole day - with different steps
		 */
		for(int i = 0; i<8; i++) { // 0 to 7 o'clock - 1 hour steps (0:00, 1:00, ... 7:00)
			long currentTS = starttime + (stepIntervall60min*i);
			//System.out.println("Appending at: " + new SimpleDateFormat("yyyyMMdd - HH:mm:ss").format(new Date(currentTS)) + " (" + currentTS + ") " + "\t Value: " + 1D/(i+1));
			fop.appendValue(label, 1D/(i+1), currentTS, (byte)i, stepIntervall60min);
		}
		//System.out.println(" ");
		for(int i = 0; i<12; i++) { // 8 to 13:30 o'clock - 30 minute steps (8:00, 8:30 ... 13:30)
			long currentTS = starttime+(8*stepIntervall60min) + (stepIntervall30min*i);
			//System.out.println("Appending at: " + new SimpleDateFormat("yyyyMMdd - HH:mm:ss").format(new Date(currentTS)) + " (" + currentTS + ") " + "\t Value: " + 1D/(i+1));
			fop.appendValue(label, 1D/(i+1), currentTS, (byte)i, stepIntervall30min);
		}
		//System.out.println(" ");
		for(int i = 0; i<8; i++) { // 14 to 15:45 o'clock - 15 minute steps (14:00, 14:15, 14:30 ... 15:45)
			long currentTS = starttime+(8*stepIntervall60min)+(12*stepIntervall30min) + (stepIntervall15min*i);
			//System.out.println("Appending at: " + new SimpleDateFormat("yyyyMMdd - HH:mm:ss").format(new Date(currentTS)) + " (" + currentTS + ") " + "\t Value: " + 1D/(i+1));
			fop.appendValue(label, 1D/(i+1), currentTS, (byte)i, stepIntervall15min);
		}
		//System.out.println(" ");
		for(int i = 0; i<8; i++) { // 16 to 23 o'clock - 1 hour steps (16:00, 17:00 ... 23:00)
			long currentTS = starttime+(8*stepIntervall60min)+(12*stepIntervall30min)+(8*stepIntervall15min) + (stepIntervall60minagain*i);
			//System.out.println("Appending at: " + new SimpleDateFormat("yyyyMMdd - HH:mm:ss").format(new Date(currentTS)) + " (" + currentTS + ") " + "\t Value: " + 1D/(i+1));
			fop.appendValue(label, 1D/(i+1), currentTS, (byte)i, stepIntervall60minagain);
		}
		
		/*
		 * File: 1293836400000.opm (1.1.2011 - 00:00)
		 */
		File toCheck = new File(SlotsDatabase.DB_TEST_ROOT_FOLDER + strDate + "/" + label + "/" + starttime + SlotsDatabase.FILE_EXTENSION);
		//SlotsDatabaseUtil.printWholeFile(toCheck);
		DataInputStream dis = new DataInputStream(new FileInputStream(toCheck));
		
		assertEquals(starttime, dis.readLong());
		assertEquals(stepIntervall60min, dis.readLong());
		
		/*
		 * Comparing every value
		 */
		for(int i = 0; i< 8; i++) {
			assertEquals(1D/(i+1), dis.readDouble());
			assertEquals((byte)i,dis.readByte());
		}
		if(dis.available() != 0)
			fail("All 8 Values are read. File should end here!");
		
		dis.close();
		
		/*
		 * File: 1293865200000.opm (1.1.2011 - 08:00)
		 */
		toCheck = new File(SlotsDatabase.DB_TEST_ROOT_FOLDER + strDate + "/" + label + "/" + (starttime+(8*stepIntervall60min)) + SlotsDatabase.FILE_EXTENSION);
		//SlotsDatabaseUtil.printWholeFile(toCheck);
		
		dis = new DataInputStream(new FileInputStream(toCheck));
		
		assertEquals(starttime+(8*stepIntervall60min), dis.readLong());
		assertEquals(stepIntervall30min, dis.readLong());
		
		/*
		 * Comparing every value
		 */
		for(int i = 0; i<12; i++) {
			assertEquals(1D/(i+1), dis.readDouble());
			assertEquals((byte)i,dis.readByte());
		}
		
		/*
		 * File: 1293886800000.opm (1.1.2011 - 14:00)
		 */
		toCheck = new File(SlotsDatabase.DB_TEST_ROOT_FOLDER + strDate + "/" + label + "/" + (starttime+(8*stepIntervall60min)+(12*stepIntervall30min)) + SlotsDatabase.FILE_EXTENSION);
		//SlotsDatabaseUtil.printWholeFile(toCheck);
		
		dis = new DataInputStream(new FileInputStream(toCheck));
		
		assertEquals(starttime+(8*stepIntervall60min)+(12*stepIntervall30min), dis.readLong());
		assertEquals(stepIntervall15min, dis.readLong());
		
		/*
		 * Comparing every value
		 */
		for(int i = 0; i<8; i++) {
			assertEquals(1D/(i+1), dis.readDouble());
			assertEquals((byte)i,dis.readByte());
		}
		
		/*
		 * File: 1293894000000.opm (1.1.2011 - 16:00)
		 */
		toCheck = new File(SlotsDatabase.DB_TEST_ROOT_FOLDER + strDate + "/" + label + "/" + (starttime+(8*stepIntervall60min)+(12*stepIntervall30min)+(8*stepIntervall15min)) + SlotsDatabase.FILE_EXTENSION);
		//SlotsDatabaseUtil.printWholeFile(toCheck);
		
		dis = new DataInputStream(new FileInputStream(toCheck));
		
		assertEquals(starttime+(8*stepIntervall60min)+(12*stepIntervall30min)+(8*stepIntervall15min), dis.readLong());
		assertEquals(stepIntervall60minagain, dis.readLong());
		
		/*
		 * Comparing every value
		 */
		for(int i = 0; i<8; i++) {
			assertEquals(1D/(i+1), dis.readDouble());
			assertEquals((byte)i,dis.readByte());
		}
	}
	
	/*
	 * Runs last test twice
	 * At second run, files already exist -> should not produce any failures
	 */
	public void testAppendValue1DayWithIntervallChangeTWICE() throws Exception {
		testAppendValue1DayWithIntervallChange();
		testAppendValue1DayWithIntervallChange();
	}
	
	/*
	 * Tests if data will be written to current file, at end of day.
	 */
	public void testWrapOverEndOfDay() throws IOException {
		long starttime = 1293836400000L; //1.1.2011 0.00 Uhr
		long stepIntervall1s = 1000; //1-second steps
		
		String label = "VoltageL1";
		String strDate1Jan2011 = new SimpleDateFormat("yyyyMMdd").format(new Date(starttime));
		String strDate2Jan2011 = new SimpleDateFormat("yyyyMMdd").format(new Date(starttime+86400000));

		/*
		 * Appending 1,5 days - with 1s steps
		 */
		for(int i = 0; i<129600; i++) { // 1.Jan 00:00:00 to 2.Jan 11:59:59 o'clock - 1 second steps
			long currentTS = starttime + (stepIntervall1s*i);
			//System.out.println("Appending at: " + new SimpleDateFormat("yyyyMMdd - HH:mm:ss").format(new Date(currentTS)) + " (" + currentTS + ") " + "\t Value: " + 1D/(i+1));
			fop.appendValue(label, i, currentTS, (byte)0x00, stepIntervall1s);
		}
		
		File toCheck1Jan = new File(SlotsDatabase.DB_TEST_ROOT_FOLDER + strDate1Jan2011 + "/" + label + "/" + starttime + SlotsDatabase.FILE_EXTENSION);
		if (!toCheck1Jan.exists()) {
			fail("File from 1.Jan 2011 missing.");
		}
		File toCheck2Jan = new File(SlotsDatabase.DB_TEST_ROOT_FOLDER + strDate2Jan2011 + "/" + label + "/" + (starttime+86400000) + SlotsDatabase.FILE_EXTENSION);
		if (!toCheck2Jan.exists()) {
			fail("File from 2.Jan 2011 missing.");
		}
		if(toCheck1Jan.getParentFile().list().length != 1) {
			fail("There should only be one file in: " + toCheck1Jan.getParentFile().getPath());
		}
		if(toCheck2Jan.getParentFile().list().length != 1) {
			fail("There should only be one file in: " + toCheck2Jan.getParentFile().getPath());
		}
		
		DataInputStream dos1 = new DataInputStream(new FileInputStream(toCheck1Jan));
		assertEquals(1293836400000L, dos1.readLong()); //comparing starttimestamp
		assertEquals(1000, dos1.readLong()); //comparing stepintervall
		
		DataInputStream dos2 = new DataInputStream(new FileInputStream(toCheck2Jan));
		assertEquals((starttime+86400000), dos2.readLong()); //comparing starttimestamp
		assertEquals(1000, dos2.readLong()); //comparing stepintervall
		
		assertEquals(86400D, dos2.readDouble()); //Checking first Value in 2.Jan file.
	}
	
	public void testAppendFor100Labels() throws IOException{
		long starttime = 1293836400000L; //1.1.2011 0.00 Uhr
		long stepIntervall60min = 3600000; //1-hour steps
		
		for(int i = 0; i<1200; i++) { // some weeks
			long currentTS = starttime + (stepIntervall60min*i);
			for (int j = 0; j<=50; j++)
				fop.appendValue("VoltageL" + j, i, currentTS, (byte)0x00, stepIntervall60min);
		}
	}
	
	public void testAppendFor1200Labels() throws IOException{
		long starttime = 1293836400000L; //1.1.2011 0.00 Uhr
		long stepIntervall60min = 3600000; //1-hour steps
		
		for(int i = 0; i<26; i++) { // about two days
			long currentTS = starttime + (stepIntervall60min*i);
			for (int j = 0; j<=1200; j++) { //1200 labels
				fop.appendValue("VoltageL" + j, i, currentTS, (byte)0x00, stepIntervall60min);
			}
		}
	}
	
	public void testReadSingleValue() throws IOException{
		long starttime = 1293836400000L; //1.1.2011 0.00 Uhr
		fop.appendValue("testLabel", 0.1, starttime+40000, (byte)State.valid.getCode(), 1000);
		fop.appendValue("testLabel", 0.2, starttime+41000, (byte)State.valid.getCode(), 1000);
		fop.appendValue("testLabel", 0.3, starttime+42000, (byte)State.valid.getCode(), 1000);
		
		fop.appendValue("testLabel", 0.4, starttime+50000, (byte)State.valid.getCode(), 5000);
		fop.appendValue("testLabel", 0.5, starttime+55000, (byte)State.valid.getCode(), 5000);
		fop.appendValue("testLabel", 0.6, starttime+60000, (byte)State.valid.getCode(), 5000);
		
		fop.appendValue("testLabel", 0.7, starttime+86401000, (byte)State.valid.getCode(), 5000);
		fop.appendValue("testLabel", 0.8, starttime+86406000, (byte)State.valid.getCode(), 5000);
		fop.appendValue("testLabel", 0.9, starttime+86411000, (byte)State.valid.getCode(), 5000);
		fop.appendValue("testLabel", 1.0, starttime+86416000, (byte)State.valid.getCode(), 5000);
		
		fop.appendValue("testLabel", 1.1, starttime+86420000, (byte)State.valid.getCode(), 3000);
		fop.appendValue("testLabel", 1.2, starttime+86423000, (byte)State.valid.getCode(), 3000);
		fop.appendValue("testLabel", 1.3, starttime+86426000, (byte)State.valid.getCode(), 3000);
		
		if (fop.read("testLabel", starttime+86417000) != null) {
			assertEquals(1.0, fop.read("testLabel", starttime+86416000).getValue());
		}
		
		if (fop.read("testLabel", starttime+41000) != null) {
			assertEquals(0.2, fop.read("testLabel", starttime+41000).getValue());
		}
	}
	
	public void testReadWholeList() throws IOException {
		long starttime = 1293836400000L; //1.1.2011 0.00 Uhr
		fop.appendValue("testLabel", 0.1, starttime+40000, (byte)State.valid.getCode(), 1000);
		fop.appendValue("testLabel", 0.2, starttime+41000, (byte)State.valid.getCode(), 1000);
		fop.appendValue("testLabel", 0.3, starttime+42000, (byte)State.valid.getCode(), 1000);
		
		fop.appendValue("testLabel", 0.4, starttime+50000, (byte)State.valid.getCode(), 5000);
		fop.appendValue("testLabel", 0.5, starttime+55000, (byte)State.valid.getCode(), 5000);
		fop.appendValue("testLabel", 0.6, starttime+60000, (byte)State.valid.getCode(), 5000);
		
		fop.appendValue("testLabel", 0.7, starttime+86401000, (byte)State.valid.getCode(), 5000);
		fop.appendValue("testLabel", 0.8, starttime+86406000, (byte)State.valid.getCode(), 5000);
		fop.appendValue("testLabel", 0.9, starttime+86411000, (byte)State.valid.getCode(), 5000);
		fop.appendValue("testLabel", 1.0, starttime+86416000, (byte)State.valid.getCode(), 5000);
		
		fop.appendValue("testLabel", 1.1, starttime+86420000, (byte)State.valid.getCode(), 3000);
		fop.appendValue("testLabel", 1.2, starttime+86423000, (byte)State.valid.getCode(), 3000);
		fop.appendValue("testLabel", 1.3, starttime+86426000, (byte)State.valid.getCode(), 3000);
		
		long somedayslater = 1294268400000L; //4 Days later.
		fop.appendValue("testLabel", 1.4, somedayslater+1000, (byte)State.valid.getCode(), 3000);
		fop.appendValue("testLabel", 1.5, somedayslater+20000, (byte)State.valid.getCode(), 3000);
		fop.appendValue("testLabel", 1.6, somedayslater+100000, (byte)State.valid.getCode(), 3000);
		
		fop.appendValue("testLabel", 1.7, somedayslater+101000, (byte)State.valid.getCode(), 1000);
		fop.appendValue("testLabel", 1.8, somedayslater+105000, (byte)State.valid.getCode(), 1000);
		fop.appendValue("testLabel", 1.9, somedayslater+109000, (byte)State.valid.getCode(), 1000);
		
		
		/**
		 * read all
		 */
		List<Value> read = fop.read("testLabel", 1293836400000L, somedayslater + 200000);
		if(read.size() != 19)
			fail("some Values are missing!");
		
		assertEquals(0.8, read.get(7).getValue());
		
		List<Value> read2 = fop.read("testLabel5", 1293836400000L, 1293836400000L);
		assertEquals(0, read2.size());
		
		List<Value> read3 = fop.read("testLabel", starttime+40100, starttime+40200);
		assertEquals(0.1, read3.get(0).getValue());
		assertEquals(1293836440000L, read3.get(0).getTimestamp());
		
		List<Value> read4 = fop.read("testLabel", starttime+44000, somedayslater+101000);
		assertEquals(14, read4.size());
		
		List<Value> read5 = fop.read("testLabel", starttime - 166400000, somedayslater+166400000);
		assertEquals(19, read5.size());
		
		List<Value> read6 = fop.read("testLabel", starttime - 166400000, starttime-86400000);
		assertEquals(0, read6.size());
		
		List<Value> read7 = fop.read("testLabel", starttime + 166400000, starttime + 266400000);
		assertEquals(0, read7.size());
		
		List<Value> read8 = fop.read("testLabel", starttime + 5000000, starttime + 5000001);
		assertEquals(0, read8.size());
		
	}

	public void testReadSomeWeeks() throws IOException {
		long starttime = 1293836400000L; //1.1.2011 0.00 Uhr
		long stepIntervall5min = 300000; //5-min- steps
		
		for(int i = 0; i<8640; i++) { // 4 weeks
			long currentTS = starttime + (stepIntervall5min*i);
				fop.appendValue("VoltageL5", i, currentTS, (byte)0x00, stepIntervall5min);
		}
		
		List<Value> read = fop.read("VoltageL5", starttime+10000000, starttime + (9000*stepIntervall5min));
		// should not raise any exceptions
		//System.out.println(read.size());
	}
	
	public void testReadWholeYear() throws IOException {
		long starttime = 1293836400000L; //1.1.2011 0.00 Uhr
		long stepIntervall5min = 300000; //5-min- steps
		
		for(int i = 0; i<10; i++) { // just some values at the beginning
			long currentTS = starttime + (stepIntervall5min*i);
				fop.appendValue("VoltageL5", i, currentTS, (byte)0x00, stepIntervall5min);
		}
		//... and some at the end of the year with different storing period...
		fop.appendValue("VoltageL5", 10.5, starttime+31535000000L, (byte)0x00, stepIntervall5min);
		fop.appendValue("VoltageL5", 10.4, starttime+31535500000L, (byte)0x00, 1000);
		fop.appendValue("VoltageL5", 10.3, starttime+31536000000L, (byte)0x00, 3000);
		
		
		List<Value> read = fop.read("VoltageL5", starttime, starttime+31536000000L); //1 whole year.
		assertEquals(13, read.size());
	}
	
	public void testDegeneratedLabels() throws IOException {
		long starttime = 1293836400000L; //1.1.2011 0.00 Uhr
		long stepIntervall5min = 300000; //5-min- steps
		
		fop.appendValue("Voltage/L1\bad,StringÄ$.", 1.0, starttime, (byte)0x00, stepIntervall5min);
		List<Value> read = fop.read("Voltage/L1\bad,StringÄ$.", 1293836400000L-10000, 1293836400000L+500);
		assertEquals(1.0, read.get(0).getValue());
	}
	
	protected void tearDown() throws Exception {
		super.tearDown();
		setUp();
	}

}
