package org.openmuc.canopen.driver;

/*
 * Copyright Fraunhofer ISE, 2011
 * Author(s): Frederic Robra
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.log4j.Logger;
import org.openmuc.canopen.exception.NoDeviceException;
import org.openmuc.canopen.exception.TimeoutException;
import org.openmuc.canopen.lib.CANopenLib;
import org.openmuc.canopen.modul.CANopenDS404Modul;
import org.openmuc.core.driver.DataReceivedListener;
import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.Measurement;
import org.openmuc.core.measurement.Unit;

/**
 * @author Frederic Robra
 *
 */
public class CANopenConnection extends Thread {
	boolean running = false;
	CANopenLib lib;
	Hashtable<Integer, CANopenDS404Modul> moduls;
	private static Logger logger = Logger.getLogger(CANopenConnection.class);
	
	private BlockingQueue<MeterReadoutTask> readoutTaskQueue;
	
	private class MeterReadoutTask {
		public DataReceivedListener callback;
		public Object callbackHandle;
		public List<ChannelIdentifier> channels;
		public DeviceURI url;

		public MeterReadoutTask(DataReceivedListener callback, Object callbackHandle,
				List<ChannelIdentifier> channels, DeviceURI url) {
			this.callback = callback;
			this.callbackHandle = callbackHandle;
			this.channels = channels;
			this.url = url;
		}
	}
	
	public CANopenConnection(String iface, String configStr) throws IOException, NoDeviceException {
		lib = new CANopenLib();
		moduls = new Hashtable<Integer, CANopenDS404Modul>();
		readoutTaskQueue = new LinkedBlockingQueue<MeterReadoutTask>();

		logger.debug("Try to connect to " + iface + "...");
		lib.connect(iface);
		logger.debug("Succesfully connected to " + iface);

	}
	
	@Override
	public void run() {
		running = true;
		
		while(running){
			try{
				if(!readoutTaskQueue.isEmpty()){
					MeterReadoutTask task = readoutTaskQueue.take();
					
					List<Measurement> data = new ArrayList<Measurement>();
					
					int modulID = new Integer(task.url.getAddress());
					if(moduls.get(modulID) == null){
						moduls.put(modulID, new CANopenDS404Modul(modulID, lib));
					}
					CANopenDS404Modul modul = moduls.get(modulID);
					long timestamp = System.currentTimeMillis();
					try {
						for(ChannelIdentifier channel : task.channels){
							int channelID = new Integer(channel.getObjectLocator());
							double value = modul.getValue(channelID);
							data.add(new Measurement(channel, value, Unit.UNKNOWN, timestamp));
						}
						task.callback.dataReceived(task.callbackHandle, data);
					}
					catch(NumberFormatException e) {
						logger.error("URL or Locator have the wrong format");
						task.callback.readoutError(task.callbackHandle, DataReceivedListener.METER_NOT_AVAILABLE);
					}
					catch(TimeoutException e) {
						logger.error(e.getMessage());
						task.callback.readoutError(task.callbackHandle, DataReceivedListener.TIMEOUT);
					}
					catch(IOException e) {
						logger.error(e.getMessage());
						task.callback.readoutError(task.callbackHandle, DataReceivedListener.INVALID_DATA);
					}
				}
				Thread.yield();
			}
			catch(InterruptedException e) { }
		}
	}

	public void close() {
		lib.close();
		logger.debug("CAN closed");
		running = false;
	}

	public void addReadoutTask(DataReceivedListener callback, Object handle,
			List<ChannelIdentifier> channels, DeviceURI url) {
		readoutTaskQueue.add(new MeterReadoutTask(callback, handle, channels, url));
	}

}
