package org.openmuc.canopen.lib;

/*
 * Copyright Fraunhofer ISE, 2011
 * Author(s): Frederic Robra
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import com.sun.jna.Library;
import com.sun.jna.Native;
import com.sun.jna.NativeLong;
import com.sun.jna.Pointer;
import com.sun.jna.Structure;
import com.sun.jna.Union;
import com.sun.jna.Structure.ByReference;

/**
 * CLibary to interact wit libc.so in Linux.
 * Implemented for communication with CAN.
 * 
 * @author Frederic Robra
 * @see JNA
 */
public interface CLibrary extends Library {
	/* Opens the native libary libc.so in Linux */
	CLibrary INSTANCE = (CLibrary)Native.loadLibrary("c", CLibrary.class);

	public static final int PF_CAN	= 29;
	public static final int SOCK_RAW = 3;
	public static final int CAN_RAW = 1;
	public static final int AF_CAN = PF_CAN;
	public static final NativeLong SIOCGIFINDEX = new NativeLong(0x8933);
	
	public static final int IFNAMESIZE = 16;
	public static final int FD_SETSIZE = 1024;
	
	/* Values for the second argument to 'fcntl'. */
	public static final int F_DUPFD = 0;					/* Duplicate file descriptor.  */
	public static final int F_GETFD = 1;					/* Get file descriptor flags.  */
	public static final int F_SETFD = 2;					/* Set file descriptor flags.  */
	public static final int F_GETFL = 3;					/* Get file status flags.  */
	public static final int F_SETFL = 4;
	public static final int O_NONBLOCK = 04000;
	
	/* Some errno variables */
	public static final int EAGAIN = 11;					/* Try again */
	public static final int EWOULDBLOCK = EAGAIN;			/* Operation would block */
	public static final int ENODEV = 19;					/* No such device */
	public static final int ENETDOWN = 100;					/* Network is down */
	
	/**
	 * Implements the C-structure sockaddr_can. <br>
	 * You need to call <pre>setType(Class type)</pre> in the can_addr union,
	 * if not called the value will remain null in the native layer.
	 */
	public class sockaddr_can extends Structure{
		public static class ByReference extends sockaddr_can implements Structure.ByReference{}
		
		public short can_family;								//unsigned short int
		public int can_ifindex;									//int
		
		public can_addrUNION can_addr;
		
		public static class can_addrUNION extends Union {
			public static class ByReference extends can_addrUNION implements Union.ByReference { }
			
			public tpSTRUCTURE tp;
			
			public static class tpSTRUCTURE extends Structure {
				public static class ByReference extends tpSTRUCTURE implements Structure.ByReference { }
				
				public int rx_id;								//unsigned int
				public int tx_id;								//unsigned int
			}
		}
	}
	
	/**
	 * Implements the C-structure sockaddr.
	 */
	public class sockaddr extends Structure {
		public static class ByReference extends sockaddr implements Structure.ByReference { }
		public short sa_family;									//unsigned shor int
		public String sa_data;
	}
	
	/**
	 * Implements the C-structure ifmap.
	 */
	public class ifmap extends Structure {
		public NativeLong mem_start;							//unsigned long int
		public NativeLong mem_end;								//unsigned long int
		public short base_addr;									//unsigned short int
		public byte irq;										//unsigned char
		public byte dma;										//unsigned char
		public byte port;										//unsigned char
	}
	
	/**
	 * Implements the C-structure ifreq. <br>
	 * You need to call
	 * <pre><code>setType(Class type)</code></pre>
	 * in the ifr_ifrn and ifr_ifru union,
	 * if not called the value will remain null in the native layer.
	 */
	public class ifreq extends Structure {
		public static class ByReference extends ifreq implements Structure.ByReference { }
		
		public ifr_ifrnUNION ifr_ifrn;
		public ifr_ifruUNION ifr_ifru;
		
		public static class ifr_ifrnUNION extends Union {
			public static class ByReference extends ifr_ifrnUNION implements Union.ByReference { }
			
			public byte[] ifrn_name = new byte[IFNAMESIZE]; 
		}
		
		public static class ifr_ifruUNION extends Union {
			public static class ByReference extends ifr_ifruUNION implements Union.ByReference { }
			
			public sockaddr ifru_addr;
			public sockaddr ifru_dstaddr;
			public sockaddr ifru_broadaddr;
			public sockaddr ifru_netmask;
			public sockaddr ifru_hwaddr;
			public short ifru_flags;
			public int ifru_ivalue;
			public int ifru_mtu;
			public ifmap ifru_map;
			public byte[] ifru_slave = new byte[IFNAMESIZE];
			public byte[] ifru_newname = new byte[IFNAMESIZE];
			public Pointer ifru_data;
			
		}
		
			
	}
	
	/**
	 * Implements the C-structure can_frame.
	 */
	public class can_frame extends Structure {
		public static class ByReference extends can_frame implements Structure.ByReference { }

		public int can_id;										//32 bit CAN_ID + EFF/RTR/ERR flags   (unsigned int)
		public byte can_dlc;									//data length code: 0 ... 8   (unsigned char)
		public byte[] none = new byte[3];						//data is aligned in a new word.
		public byte[] data = new byte[8];						//unsigned char data[8] __attribute__((aligned(8)));
	}
	
	/**
	 * Implements the C-structure fd_set.
	 */
	public class fd_set extends Structure {
		public static class ByReference extends fd_set implements Structure.ByReference { }
		
		public fd_set() {
			for(int i = 0; i < __fds_bits.length; i++){
				__fds_bits[i] = new NativeLong(0);
			}
		}
		/* __fds_bits.length is 32 with 32Bit and 16 with 64Bit */
		public NativeLong[] __fds_bits = new NativeLong[FD_SETSIZE / (8 * NativeLong.SIZE)];
		/**
		 * <b>#define</b> FD_SET(fd, fdsetp)
		 * <pre><code>
		 * (((fdsetp)->__fds_bits)[((fd) / (8 * sizeof(unsigned long)))] |= (1UL << ((fd) % (8 * sizeof(unsigned long)))))
		 * </code></pre>
		 * @param fd File descriptor
		 */
		public void set(int fd){
			long fd_bit = __fds_bits[(fd / (8 * NativeLong.SIZE))].longValue();
			fd_bit |= (1l << ((fd) % (8 * NativeLong.SIZE)));
			__fds_bits[(fd / (8 * NativeLong.SIZE))].setValue(fd_bit);
		}
	}
	
	/**
	 * Implements the C-structure timeval.
	 */
	public class timeval extends Structure {
		public static class ByReference extends timeval implements Structure.ByReference { }
		
		public NativeLong tv_sec = new NativeLong(0);
		public NativeLong tv_usec = new NativeLong(0);
	}
	
	/**
	 * Create a new socket of type TYPE in domain DOMAIN, using
	 * protocol PROTOCOL.  If PROTOCOL is zero, one is chosen automatically.
	 * Returns a file descriptor for the new socket, or -1 for errors.
	 * @param __domain DOMAIN
	 * @param __type TYPE
	 * @param __protocol PROTOCOL If PROTOCOL is zero, one is chosen automatically.
	 * @return A file descriptor for the new socket, or -1 for errors
	 */
	public int socket(int __domain, int __type, int __protocol);
	
	/**
	 * Perform the I/O control operation specified by REQUEST on FD.
	 * One argument may follow; its presence and type depend on REQUEST.
	 * Return value depends on REQUEST.  Usually -1 indicates error.
	 * @param __fd File descriptor
	 * @param __request REQUEST Specifies the operation
	 * @param objects Varargs
	 * @return Value depends on REQUEST. Usually -1 indicates error.
	 */
	public int ioctl(int __fd, NativeLong __request, Object...objects);
	
	/**
	 * Give the socket FD the local address ADDR (which is LEN bytes long).
	 * @param __fd File descriptor
	 * @param __addr The local address ADDR.
	 * @param __len The size of ADDR in bytes.
	 * @return -1 for errors.
	 */
	public int bind(int __fd, ByReference __addr, int __len);
	
	/**
	 * Close the file descriptor FD.
	 * @param __fd File descriptor
	 * @return -1 for errors.
	 */
	public int close(int __fd);
	
	/**
	 * Check the first NFDS descriptors each in READFDS (if not NULL) for read
	 * readiness, in WRITEFDS (if not NULL) for write readiness, and in EXCEPTFDS
	 * (if not NULL) for exceptional conditions.  If TIMEOUT is not NULL, time out
	 * after waiting the interval specified therein.  Returns the number of ready
	 * descriptors, or -1 for errors.
	 * @param __nfds NFDS
	 * @param __readfds READFDS
	 * @param __writefds WRITEFDS
	 * @param __exceptfds EXCEPTFDS
	 * @param __timeout TIMEOUT
	 * @return  The number of ready descriptors, 0 if timeout, -1 if error.
	 */
	public int select (int __nfds, fd_set.ByReference __readfds, fd_set.ByReference __writefds,
						fd_set.ByReference __exceptfds, timeval.ByReference __timeout);
	/**
	 * Do the file control operation described by CMD on FD.
	 * The remaining arguments are interpreted depending on CMD.
	 * @param __fd File descriptor
	 * @param __cmd CMD
	 * @param objects Varargs
	 * @return
	 */
	public int fcntl (int __fd, int __cmd, Object...objects);
	
	/**
	 * Write N bytes of BUF to FD.  Return the number written, or -1.
	 * @param __fd File descriptor
	 * @param __buf BUF 
	 * @param __nbytes Bytes to be written.
	 * @return The number written, or -1.
	 */
	public int write(int __fd, Object __buf, int __nbytes);
	
	/**
	 * Read NBYTES into BUF from FD.  Return the
	 * number read, -1 for errors or 0 for EOF.
	 * @param __fd File descriptor
	 * @param __buf BUF
	 * @param __nbytes Bytes to be read.
	 * @return The number read, -1 for errors or 0 for EOF.
	 */
	public int read(int __fd, Object __buf, int __nbytes);
	
	/**
	 * Write formatted output to stdout.
	 * @param format Formatted string.
	 * @param objects Varargs to be written.
	 */
	public void printf(String format, Object...objects);
	
	/**
	 * Return a string describing the meaning of the 'errno' code in ERRNUM.
	 * @param errnum errno
	 * @return A string describing the meaning of the 'errno' code in ERRNUM.
	 */
	public String strerror(int errnum);
	
	/**
	 * Copy SRC to DEST.
	 * @param __dest DEST
	 * @param __src SRC
	 * @return The String copied.
	 */
	public String strcpy(byte[] __dest, String __src);
	
}
