package org.openmuc.ehz.sml;

import gnu.io.PortInUseException;
import gnu.io.UnsupportedCommOperationException;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import org.openmuc.core.measurement.Unit;
import org.openmuc.core.obis.OBISCode;
import org.openmuc.ehz.driver.Device;
import org.openmuc.ehz.driver.DeviceInfo;
import org.openmuc.ehz.exception.NotFoundException;
import org.openmuc.jsml.structures.ASNObject;
import org.openmuc.jsml.structures.Integer16;
import org.openmuc.jsml.structures.Integer32;
import org.openmuc.jsml.structures.Integer64;
import org.openmuc.jsml.structures.Integer8;
import org.openmuc.jsml.structures.SML_File;
import org.openmuc.jsml.structures.SML_GetListRes;
import org.openmuc.jsml.structures.SML_ListEntry;
import org.openmuc.jsml.structures.SML_Message;
import org.openmuc.jsml.structures.SML_MessageBody;
import org.openmuc.jsml.structures.Unsigned16;
import org.openmuc.jsml.structures.Unsigned32;
import org.openmuc.jsml.structures.Unsigned64;
import org.openmuc.jsml.structures.Unsigned8;
import org.openmuc.jsml.tl.SML_SerialReceiver;


public class SMLDevice extends Device{

	private SML_SerialReceiver receiver;
	private String serverID;
	
	public SMLDevice(String iface, String config) throws NotFoundException, IOException {
		super(iface, config);

		receiver = new SML_SerialReceiver();
		try {
			receiver.setupComPort(iface);
		}
		catch (PortInUseException e) {
			throw new NotFoundException("Port in use");
		}
		catch (UnsupportedCommOperationException e) {
			throw new NotFoundException("Unsupported comm operation");
		}
	}

	@Override
	public Hashtable<String, Double> getValues() throws IOException {
		Hashtable<String, Double> values = new Hashtable<String, Double>();
		
		SML_ListEntry[] list = getSML_ListEntries();
				
		for (SML_ListEntry entry : list) {
			String address = getAddress(entry.getObjName().getOctetString());
			values.put(address, getValue(entry));
		}

		return values;
	}

	@Override
	public List<DeviceInfo> getDeviceInfos() throws IOException {
		List<DeviceInfo> infos = new ArrayList<DeviceInfo>();
		SML_ListEntry[] list = getSML_ListEntries();
		for (SML_ListEntry entry : list) {
			DeviceInfo info = new DeviceInfo();
			String address = getAddress(entry.getObjName().getOctetString());
			info.setAddress(address);
			info.setValue(getValue(entry));
			info.setUnit(Unit.fromUnitCode(entry.getUnit().getVal()));
			String description = (Unit.fromUnitCode(entry.getUnit().getVal()).name());
			info.setDescription(description);
			
			infos.add(info);
		}
		return infos;
	}
	
	private SML_ListEntry[] getSML_ListEntries() throws IOException {
		SML_File smlFile = receiver.getSMLFile();

		List<SML_Message> messages = smlFile.getMessages();
		
		SML_ListEntry[] list = null;
		
		for (SML_Message message : messages) {
			int tag = message.getMessageBody().getTag().getVal();

			if (tag == SML_MessageBody.GetListResponse) {
				SML_GetListRes resp = (SML_GetListRes) message.getMessageBody().getChoice();
				
				if (serverID == null) {
					serverID = "";
					for (Byte b : resp.getServerId().getOctetString()) {
						serverID += Integer.toString((b & 0xff) + 0x100, 16).substring(1);
					}
					serverID = serverID.toUpperCase();
				}

				list = resp.getValList().getValListEntry();
				break;
			}
		}
		return list;
	}
	
	private double getValue(SML_ListEntry entry) throws IOException {
		double value = 0;

		ASNObject obj = entry.getValue().getChoice();
		if (obj.getClass().equals(Integer64.class)) {
			Integer64 val = (Integer64) obj;
			value = val.getVal();
		}
		else if (obj.getClass().equals(Integer32.class)) {
			Integer32 val = (Integer32) obj;
			value = val.getVal();
		}
		else if (obj.getClass().equals(Integer16.class)) {
			Integer16 val = (Integer16) obj;
			value = val.getVal();
		}
		else if (obj.getClass().equals(Integer8.class)) {
			Integer8 val = (Integer8) obj;
			value = val.getVal();
		}
		else if (obj.getClass().equals(Unsigned64.class)) {
			Unsigned64 val = (Unsigned64) obj;
			value = val.getVal();
		}
		else if (obj.getClass().equals(Unsigned32.class)) {
			Unsigned32 val = (Unsigned32) obj;
			value = val.getVal();
		}
		else if (obj.getClass().equals(Unsigned16.class)) {
			Unsigned16 val = (Unsigned16) obj;
			value = val.getVal();
		}
		else if (obj.getClass().equals(Unsigned8.class)) {
			Unsigned8 val = (Unsigned8) obj;
			value = val.getVal();
		}
		else {
			return Double.NaN;
		}
		
		byte scaler = entry.getScaler().getVal();
		return value * Math.pow(10, scaler);
	}

	private String getAddress(byte[] data) {
		OBISCode code = new OBISCode(data);
		return code.toString();
	}
	
	@Override
	public String getID() {
		return serverID;
	}
	
	@Override
	public String getName() {
		return "eHZ SML";
	}
	
	@Override
	public void close() {
		try {
			receiver.close();
		}
		catch (IOException e) { }
	}

}
