package org.openmuc.modbus;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Simon Fey
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.List;
import java.util.Vector;

import net.wimpi.modbus.procimg.InputRegister;

import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.driver.MethodNotImplementedException;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.Measurement;
import org.openmuc.core.measurement.Unit;
import org.openmuc.core.measurement.Value;

public class ModbusDevice {
	private ModbusConnection connection;
	private DeviceURI url = null;
	
	public ModbusDevice(DeviceURI url) throws Exception {
		this(url.getAddress());
		this.url = url;
	}
	
	public ModbusDevice(String serverIP) throws Exception {
		this.connection = new ModbusConnection(serverIP);
	}
	
	public DeviceURI getUrl() {
		return url;
	}
	
	public void disconnect() {
		this.connection.close();
	}
	
	public void writeOutputRegister(ChannelIdentifier register, double value) 
		throws MethodNotImplementedException
	{
		try {
			ModbusObjectLocator mol = new ModbusObjectLocator(register.getObjectLocator());
			this.connection.writeOutputRegister(mol.getRegisterAddress(), (int) value); //TODO: cast von double auf int? oder eher short?
			if (Boolean.valueOf(System.getProperty("org.openmuc.modbus.debug")).booleanValue()) {
				System.out.println("MODBUS DEBUG: Writing device: " + this.getUrl() + " - " + register.getLabel() + " (register: " + mol.getRegisterAddress() + ", type: " + mol.getRegisterDataType() + "): " + value);
			}
		} catch (MalformedObjectLocatorException e) {
			e.printStackTrace();
		}
	}
	
	public Vector<Measurement> getChannels(List<ChannelIdentifier> registers) 
		throws Exception
	{
		long timestamp = System.currentTimeMillis();
		InputRegister[] inRegs = null;
		Vector<Measurement> measurements = new Vector<Measurement>();
		
		for (ChannelIdentifier register : registers) {
			String locator = register.getObjectLocator();
			ModbusObjectLocator mol = new ModbusObjectLocator(locator);
			double meterVal = 0; //This double contains the Register Content from the Modbus Device.
			
			inRegs = this.connection.readInputRegisters(mol.getRegisterAddress(), mol.getRegisterCount());
			/* inRegs:
			 * Modbus Devices may use several sequenced Registers to store Data.
			 * The Register Values must be concatenated for further processing.
			 */ 
			if (inRegs[0].toBytes().length > 0) {
				
				byte[] concate = new byte[inRegs.length*inRegs[0].toBytes().length];
	
				for (int i=0; i<inRegs.length; i++) {
					System.arraycopy(inRegs[i].toBytes(), 0, concate, i*inRegs[i].toBytes().length, inRegs[i].toBytes().length);
				}
				
				ByteBuffer buf;
				buf = ByteBuffer.wrap(concate);
				buf.position(0);
				buf.order(ByteOrder.BIG_ENDIAN);
				
				/* DataType:
				 * buf does now contain a Byte Buffer, which has to be interpreted by it's correct Data Type.
				 * (int, float, double, short e.g.)
				 * As a reason of storing the data in a DB, the final data type will always be a double.
				 */
				
				switch (mol.getRegisterDataType()) {
				case dt_float:
					float f = buf.getFloat();
					meterVal = (double) f; 
					break;
				case dt_double:
					meterVal = buf.getDouble();
					break;
				case dt_int:
					int i = buf.getInt();
					meterVal = (double) i;
					break;
				case dt_short:
					short s = buf.getShort();
					meterVal = (double) s;
					break;
				case dt_long:
					long l = buf.getLong();
					meterVal = (double) l;
					break;
				}
			} else {
				//TODO: exception necessary?
				throw new Exception("InputRegisters empty");
			}
			
			if (Boolean.valueOf(System.getProperty("org.openmuc.modbus.debug")).booleanValue()) {
				System.out.println("MODBUS DEBUG: Reading device: " + this.getUrl() + " - " + register.getLabel() + " (register: " + mol.getRegisterAddress() + ", type: " + mol.getRegisterDataType().toString().substring(3) + "): " + meterVal);
			}
				
			Value val = new Value(meterVal, timestamp);
			Measurement meas = new Measurement(register, val, Unit.UNKNOWN); //TODO: unit?!
			measurements.add(meas);
		}
		return measurements;
	}
	
}
