package org.openmuc.metering.driver.s7plc;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.com.libnodave.Connection;
import org.openmuc.core.driver.DataReceivedListener;
import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.measurement.Measurement;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.Unit;
import org.openmuc.core.measurement.Value;
import org.openmuc.metering.driver.s7plc.S7PLCDriverImpl.PLCConnection;

public class ReadoutThread implements Runnable {

	private Logger logger = Logger.getLogger(ReadoutThread.class);
	
	private S7PLCDriverImpl driver;

	private List<ChannelIdentifier> channels;
	
	private DataReceivedListener callback;
	
	private Object handle;

	private class DataBlock {
		int nr;
		int maxOffset = 0;
		byte[] buf = null;
		
		public DataBlock(int nr) {
			this.nr = nr;
		}
	}
	
	private class DataObject {
		ChannelIdentifier mdi;
		S7ObjectLocator loc;
		
		public DataObject(ChannelIdentifier mdi, S7ObjectLocator loc) {
			this.mdi = mdi;
			this.loc = loc;
		}
	}
	
	private HashMap<Integer,DataBlock> dbList;
	private LinkedList<DataObject> objList;
	private DeviceURI url;
	
	public ReadoutThread(S7PLCDriverImpl driver, List<ChannelIdentifier> registers2,
			DataReceivedListener callback, Object handle, DeviceURI url) 
	{
		super();
		this.channels = registers2;
		this.callback = callback;
		this.handle = handle;
		this.driver = driver;
		this.url = url;
	}
	
	private byte[] concat(byte[] A, byte[] B) {
		byte[] C = new byte[A.length + B.length];
		System.arraycopy(A, 0, C, 0, A.length);
		System.arraycopy(B, 0, C, A.length, B.length);

		return C;
	}

	
	public void run() {
		
		try {
			Vector<Measurement> dos = new Vector<Measurement>();
			
			/* Prepare list of data blocks to read out */
			dbList = new HashMap<Integer,DataBlock>();
			objList = new LinkedList<DataObject>();

			PLCConnection con = this.driver.getConnection(this.url);
			
			for (ChannelIdentifier dataID : channels) {

				try {
					S7ObjectLocator loc = new S7ObjectLocator(dataID.getObjectLocator());
					
					int dbAddr = loc.getAreaAddress();
					DataBlock db;
					
					if (!dbList.containsKey(dbAddr)) {
						dbList.put(dbAddr, new DataBlock(dbAddr));
					}
					
					db = dbList.get(dbAddr);
					
					int size = loc.getOffset() + loc.getDataLength();
					
					if (size > db.maxOffset) db.maxOffset = size;
					
					objList.add(new DataObject(dataID, loc));
					
				} catch (MalformendObjectLocatorException e) {
					System.err.println("MaformedObjectLocator: " + dataID.getObjectLocator());
					callback.readoutError(handle, 2);
					return;
				}
				
			} /* for */
			
			/* Read out data blocks */
			for (DataBlock db : dbList.values()) {
				// synchronized (con) {
				
					if (con != null) {
										
						try {
							
							int maxOffset = db.maxOffset;
							
							int startOffset = 0;
							
							while (maxOffset > 460) {
								if (db.buf == null) {
									db.buf = con.con.readBytes(Connection.AREA_DB, db.nr, startOffset, 460);
								}
								else {
									db.buf = concat(db.buf, con.con.readBytes(Connection.AREA_DB, db.nr, startOffset, 460));
								}
								startOffset += 460;
								maxOffset -= 460;
							}

							if (db.buf == null) {
								db.buf = con.con.readBytes(Connection.AREA_DB, db.nr, startOffset, maxOffset);
							}
							else {
								db.buf = concat(db.buf, con.con.readBytes(Connection.AREA_DB, db.nr, startOffset, maxOffset));
							}
							
						}
						catch (IOException e) {
							logger.info("plcs7: readout error - close socket");
							
							con.ifc.close();
							driver.removeConnection(con);
							
							callback.readoutError(handle, 2);
							return;
						}
					} 
					else {
						callback.readoutError(handle, 2);
						return;
					}
				// }
			} /* for */
			
			/* Get data out of buffers */
			for (DataObject obj : objList) {
				DataBlock db = dbList.get(obj.loc.getAreaAddress());
				
				if (db.buf != null) {
					double val = 0.0;
					ByteBuffer bbuf;
					bbuf = ByteBuffer.wrap(db.buf);
					bbuf.order(ByteOrder.BIG_ENDIAN);

					switch (obj.loc.getType()) {
					case S7ObjectLocator.TYPE_FLOAT:
						val = (double) bbuf.getFloat(obj.loc.getOffset());
						break;
					case S7ObjectLocator.TYPE_UINT32:
						val = (double) ((long) 0xffffffff & (long) bbuf.getInt(obj.loc.getOffset()));
						break;
					case S7ObjectLocator.TYPE_UINT16:
						val = (double) (0xffff & (int) bbuf.getShort(obj.loc.getOffset()));
						break;
					case S7ObjectLocator.TYPE_UINT8:
						val = (double) (0xff & (int) bbuf.get(obj.loc.getOffset()));
						break;
					case S7ObjectLocator.TYPE_BIT:
						val = (double) getBit(bbuf.get(obj.loc.getOffset()), obj.loc.getBitPos());				
						break;
					}

					Value value = new Value(val, System.currentTimeMillis());

					Measurement meas = new Measurement(obj.mdi, value, Unit.UNKNOWN);
					dos.add(meas);

				} else {
					callback.readoutError(handle, 2);
					return;
				}
				
				
			}
			
			callback.dataReceived(handle, dos);
		} catch (Exception e) {			
			/* Log error stack trace */
			OutputStream outs = new ByteArrayOutputStream();
			PrintWriter pw = new PrintWriter(outs);
			e.printStackTrace(pw);
			pw.flush();
			
			logger.error(outs.toString());
			
			callback.readoutError(handle, 2);
		}
	} /* run() */

	private int getBit(byte b, byte bitPos) {
		switch (bitPos) {
		case 0:
			if ((b & 0x01) != 0) return 1; else return 0;
		case 1:
			if ((b & 0x02) != 0) return 1; else return 0;
		case 2:
			if ((b & 0x04) != 0) return 1; else return 0;
		case 3:
			if ((b & 0x08) != 0) return 1; else return 0;
		case 4:
			if ((b & 0x10) != 0) return 1; else return 0;
		case 5:
			if ((b & 0x20) != 0) return 1; else return 0;
		case 6:
			if ((b & 0x40) != 0) return 1; else return 0;
		case 7:
			if ((b & 0x80) != 0) return 1; else return 0;
		default:
			return 0;
		}
	} /* getBit() */

}
