package org.openmuc.sunnywebbox;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Simon Fey
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;

import org.apache.log4j.Logger;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.JSONValue;
import org.json.simple.parser.JSONParser;
import org.json.simple.parser.ParseException;
import org.openmuc.core.driver.BusURI;
import org.openmuc.core.driver.ChannelInformation;
import org.openmuc.core.driver.DataReceivedListener;
import org.openmuc.core.driver.DeviceDriverInterface;
import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.driver.MethodNotImplementedException;
import org.openmuc.core.driver.NoSuchInterfaceException;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.Measurement;

public class SunnyWebboxDriverImpl implements DeviceDriverInterface{
	private static int instanceCount = 0;
	private int instanceID = 0;
	private int threadCount = 0;
	public boolean debug = false;
	
	private Vector<SunnyWebboxDevice> devicePool = new Vector<SunnyWebboxDevice>();
	private Executor executor;
	
	private Map<String, Object> json_map;
	private JSONObject json_obj;
	private JSONParser json_parser;
	private JSONArray json_array;
	
	private SunnyWebboxConnection connection;
	
	private static Logger logger = Logger.getLogger(SunnyWebboxDriverImpl.class);
	
	public SunnyWebboxDriverImpl() {
		json_map = new LinkedHashMap();
		json_parser = new JSONParser();
		
		instanceID = instanceCount++;
		executor = Executors.newCachedThreadPool(new ThreadFactory() {
			public Thread newThread(Runnable arg0) {
				return new Thread(arg0, "sunnwebboxthreadpool-" + instanceID + "-" + threadCount++);
			}
		});
	}
	
	public void configureInterface(String iface, String configStr)
			throws NoSuchInterfaceException, MethodNotImplementedException,
			IOException {
		// TODO Auto-generated method stub
	}

	public List<DeviceURI> getDeviceList(BusURI url)
			throws MethodNotImplementedException, IOException {

		List<DeviceURI> toReturn = new Vector<DeviceURI>();
		
		this.connection = new SunnyWebboxConnection(url.getComPort());
		
		List<String> deviceStrings = new Vector<String>();
		
		json_map.put("version", "1.0");
		json_map.put("proc", "GetDevices");
		json_map.put("id", "1");
		json_map.put("format", "JSON");
		String response = connection.sendJSONRequest(JSONValue.toJSONString(json_map));
		if (response != null) {
			try {
				json_obj = (JSONObject) json_parser.parse(response);
			} catch (ParseException e) {
				logger.error("Unable to Parse JSON Response from Sunny Webbox");
				e.printStackTrace();
			}
			json_obj = (JSONObject) json_obj.get("result");
			json_array = (JSONArray) json_obj.get("devices");
			ListIterator<JSONObject> iter = json_array.listIterator();
			while (iter.hasNext()) {
				json_obj = (JSONObject) iter.next();
				deviceStrings.add((String) json_obj.get("key"));
			}
		}
		for (String device : deviceStrings) {
			toReturn.add(new DeviceURI(url.toString() + ":" + device));
		}
		
		return toReturn;
	}

	public String getDriverID() {
		return "sunnywebbox";
	}

	public List<String> getSupportedOperations() {
		// TODO Auto-generated method stub
		return null;
	}

	public void readDeviceAsync(DeviceURI url,
			List<ChannelIdentifier> channels, DataReceivedListener callback,
			Object handle) throws MethodNotImplementedException {
		SunnyWebboxDevice dev = null;
		
		for(SunnyWebboxDevice device : devicePool) {
			if (device.getUrl().getAddress().equals(url.getAddress())) { //compare also to comPort?!
				dev = device;
				break;
			}
		}
		
		/* Add new device to device pool */
		if (dev == null) {
			try {
				dev = new SunnyWebboxDevice(url);
				devicePool.add(dev);
			} catch (Exception e) {
				logger.error("Cannot connect to " + url.toString());
				return;
			}
		}
		
		executor.execute(new ReadoutTask(dev, channels, callback, handle));
	}

	public List<Measurement> readDeviceSync(DeviceURI url,
			List<ChannelIdentifier> channels)
			throws MethodNotImplementedException, IOException {
		
		SunnyWebboxDevice dev = null;
		List<Measurement> toReturn = new Vector<Measurement>();
		
		for(SunnyWebboxDevice device : devicePool) {
			if (device.getUrl().getAddress().equals(url.getAddress())) {
				dev = device;
				break;
			}
		}
		
		/* Add new device to device pool */
		if (dev == null) {
			try {
				dev = new SunnyWebboxDevice(url);
				devicePool.add(dev);
			} catch (Exception e) {
				logger.error("Cannot connect to " + url.toString());
				return null;
			}
		}
		
		try {
			toReturn = dev.getMeasurements(channels);
		} catch (MalformedObjectLocatorException e) {
			logger.error("Malformed Object Locator for Sunny Webbox Driver.");
		}
		
		return toReturn;
	}

	public void writeChannel(DeviceURI url, ChannelIdentifier channel,
			double value) throws MethodNotImplementedException, IOException {
		throw new MethodNotImplementedException();
	}

	public List<ChannelInformation> getDeviceChannels(DeviceURI url)
			throws MethodNotImplementedException, IOException {
		
		List<ChannelInformation> channels = new Vector<ChannelInformation>();
		
		SunnyWebboxDevice dev = null;
		
		for(SunnyWebboxDevice device : devicePool) {
			if (device.getUrl().getComPort().equals(url.getComPort()) && // ComPort is IP-Address
					device.getUrl().getAddress().equals(url.getAddress())) { // Address is Name of a WR attached to the Webbox
				dev = device;
				break;
			}
		}
		
		/* Add new device to device pool */
		if (dev == null) {
			try {
				dev = new SunnyWebboxDevice(url);
				devicePool.add(dev);
			} catch (Exception e) {
				e.printStackTrace();
				System.err.println("Cannot connect to " + url.toString());
				return null;
			}
		}
		
		try {
			channels = dev.getChannelsForDevice(url.getAddress());
		} catch (ParseException e) {
			e.printStackTrace();
			throw new IOException("Error while parsing JSON Response from webbox");
		}
		
		return channels;
	}
	
	public void disconnect() {
		connection.close();
	}
	
}
