package org.openmuc.mux.dataserver;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.openmuc.core.datamanager.Authentication;
import org.openmuc.core.datamanager.DataManager;
import org.openmuc.core.datamanager.DataStorage;
import org.openmuc.core.measurement.Value;

/**
 * 
 * Client connection handler. For each connected client a Connection object will
 * be created. Handles the client requests.<br>
 * 
 * The following commands are currently supported: <li>
 * ?org.openmuc.core.datamanager.dataserver <li>?get [label] <li>?get_age [label] <li>
 * ?get_value <li>?set_value <li>?get_values <li>?configure_report <li>
 * ?get_directory <li>?disconnect
 * 
 * <br>
 * <br>
 * FIXME Reporting doesn't work with the current version!
 * 
 * @author mzillgit
 * 
 */
public class Connection extends Thread {

	private static Logger logger = Logger.getLogger(Connection.class);
	
	private DataManagerServiceTracker store;
	private Socket sock;
	private InputStream input;
	private OutputStream output;
	private BufferedReader br;
	private PrintWriter out;
	private Authentication auth;
	private DataSocketServer myServer;
	private boolean running = true;
	private Vector<Report> reports = new Vector<Report>();

	private static int reportHandle = 0;

	public Connection(DataManagerServiceTracker store, Authentication auth,
			Socket sock, DataSocketServer server) throws IOException {
		this.store = store;
		this.sock = sock;
		this.myServer = server;
		this.auth = auth;

		this.input = sock.getInputStream();
		this.output = sock.getOutputStream();

		this.br = new BufferedReader(new InputStreamReader(input));
		this.out = new PrintWriter(output, true);
		this.sock.setSoTimeout(60000);

	}

	public void close() {
		// TODO ?`Wofuer wird diese Funktion verwendet ?
		try {
			this.sock.close();
		} catch (IOException e) {
			logger.error(e.getMessage());
		}
		// this.interrupt();
	}

	@Override
	public void run() {
		String msg;

		while (!this.sock.isInputShutdown() && (this.running == true)) {

			try {
				Thread.sleep(100);
			} catch (InterruptedException e1) {
				e1.printStackTrace();
			}

			try {
				msg = br.readLine();
				if (msg != null) {
					logger.info("REQUEST: " + msg);

					if (msg.startsWith("?org.openmuc.mux.dataserver")) {
						out.println("!org.openmuc.mux.dataserver_response");
					} else if (msg.startsWith("?get ")) {
						getReply(msg);
					} else if (msg.startsWith("?get_age ")) {
						getAge(msg);
					} else if (msg.startsWith("?get_value ")) {
						getValueReply(msg);
					} else if (msg.startsWith("?set_value ")) {
						setValueReply(msg);
					} else if (msg.startsWith("?get_all_values")) {
						getAllValuesReply(msg);
					} else if (msg.startsWith("?get_values ")) {
						getValuesReply(msg);
					} else if (msg.startsWith("?configure_report ")) {
						configureReport(msg);
					} else if (msg.startsWith("?get_directory")) {
						getDirectoryReply(msg);
					} else if (msg.startsWith("?disconnect")) {
						out.println("!DISCONNECTED");
						this.sock.close();
						this.running = false;
					} else {
						out.println("!ERROR");
					}
				} else {
					logger.info("CLOSE SOCKET");
					this.sock.close();
					this.running = false;
				}
			} catch (InterruptedIOException iioe) {
				try {
					this.sock.close();
				} catch (IOException e) {
					logger.error(e.getMessage());
				}
				logger.info("Timeout occurred (60s) - killing connection");
				break;
			} catch (Exception e) {
				logger.info(e.getMessage());
				this.running = false;
			}

			/* create reports on demand */
			// TODO Implement reporting with timer / separate thread
			// if (store.getStore() != null) {
			// for (Report report : reports) {
			// report.emitReport(out);
			// }
			// }
			// else {
			// if (reports.size() > 0) reports.clear();
			// }

		}

		logger.info("CONNECTION CLOSED!");
		this.myServer.connectionClosed(this);
	} /* run() */

	private void getAllValuesReply(String msg) {
		if (store.getManager() != null) {
			List<String> directory = this.store.getManager().getDeviceDirectory();

			out.print("!GET_ALL_VALUES_RESPONSE");

			for (String label : directory) {
				DataStorage storage = store.getManager().getDataStorage(label, auth);
			
				if (storage != null) {
					if (storage.getCurrentValue() != null)
						out.print(" " + label + ":" + storage.getCurrentValue().getValue());
					else
						out.print(" " + label + ":null");
				}
			}

			out.println();
		} else
			out.println("!ERROR: data store service not present!");
	}

	private void setValueReply(String msg) {
		StringTokenizer st = new StringTokenizer(msg);
		String label, value;

		st.nextToken(); // Skip command
		label = st.nextToken();
		value = st.nextToken();

		if ((label == null) || (value == null)) {
			out.println("!ERROR: argument missing!");
			return;
		}

		DataStorage storage = store.getManager().getDataStorage(label, auth);

		if (storage != null) {

			double dblVal = Double.valueOf(value);

			storage.setValue(dblVal);
		} else {
			out.println("!ERROR: label not found!");
			return;
		}

		out.println("!OK");
	} /* setValueReply() */

	private void getAge(String msg) {
		StringTokenizer st = new StringTokenizer(msg);
		String label;

		try {
			st.nextToken(); // Skip command

			label = st.nextToken();

			logger.debug("?get_age label: (" + label + ")");

			if (store.getManager() != null) {
				long lastUpdate;
				long age;

				DataStorage storage = store.getManager().getDataStorage(label,
						auth);

				lastUpdate = storage.getLastUpdateTime();

				age = System.currentTimeMillis() - lastUpdate;

				out.println("!GET_AGE_RESPONSE " + label + " " + age + " ms");
			} else
				out.println("!ERROR: data store service not present!");
		} catch (Exception e) {
			out.println("!ERROR");
		logger.error(e.getMessage());
		}
	} /* getAge() */

	private void configureReport(String msg) {
		StringTokenizer st = new StringTokenizer(msg);

		try {
			Report report;
			String type;
			
			st.nextToken(); // Skip command

			type = st.nextToken();

			if (type.equals("periodic")) {
				String interval = st.nextToken();

				int intvl = Integer.parseInt(interval);

				report = new PeriodicReport(reportHandle++, intvl);
			} else {
				out.println("!ERROR: Unknown report type!");
				return;
			}

			String label;

			DataManager storeSvc = store.getManager();

			if (storeSvc != null) {

				while (st.hasMoreTokens()) {
					label = st.nextToken();
					DataStorage storage = storeSvc.getDataStorage(label, auth);

					if (storage != null)
						report.addDataStorage(storage);
					else {
						out.println("!ERROR: label not known!");
						return;
					}
				}
			} else
				out.println("!ERROR: data store service not present!");

			reports.add(report);

			out.println("!CONFIGURE_REPORT_RESPONSE handle:"
					+ report.getHandle());
		} catch (Exception e) {
			out.println("!ERROR");
			logger.error(e.getMessage());
		}
	}

	private void getReply(String msg) {
		StringTokenizer st = new StringTokenizer(msg);
		String label;

		try {
			st.nextToken(); // Skip command

			label = st.nextToken();

			logger.debug("?get label: (" + label + ")");

			if (store.getManager() != null) {

				DataStorage storage = store.getManager().getDataStorage(label,
						auth);

				Value value = storage.getCurrentValue();

				if (value != null) {
					out.println("!GET_RESPONSE " + label + " "
							+ value.getValue());
				} else
					out.println("!ERROR: Value not found!");
			} else
				out.println("!ERROR: data store service not present!");
		} catch (Exception e) {
			out.println("!ERROR");
			logger.error(e.getMessage());
		}
	} /* getReply() */

	private void getValueReply(String msg) {
		StringTokenizer st = new StringTokenizer(msg);
		String label;
		String timestamp;

		try {
			st.nextToken(); // Skip command

			label = st.nextToken();
			timestamp = st.nextToken();

			logger.debug("?get_value label: (" + label + ") at time " + timestamp);

			if (store.getManager() != null) {

				DataStorage storage = store.getManager().getDataStorage(label,
						auth);

				Value value = storage.getValue(Long.parseLong(timestamp));

				if (value != null) {
					out.println("!GET_VALUE_RESPONSE " + label + " "
							+ value.getValue());
				} else
					out.println("!ERROR: Value not found!");
			} else
				out.println("!ERROR: data store service not present!");
		} catch (Exception e) {
			out.println("!ERROR");
			logger.error(e.getMessage());
		}

	}

	private void getValuesReply(String msg) {
		StringTokenizer st = new StringTokenizer(msg);
		String label;
		String startTime;
		String endTime;

		try {
			st.nextToken(); // Skip command

			label = st.nextToken();

			startTime = st.nextToken();

			logger.debug("?get_values label: (" + label + ") at time " + startTime);

			if (store.getManager() != null) {

				DataStorage storage = store.getManager().getDataStorage(label,
						auth);

				List<Value> values;
				
				System.out.println("got storage for label: " + label);

				if (storage == null) System.out.println("Storage == null !!!");
				
				System.out.println("storage: " + storage.toString());
				
				if (st.hasMoreTokens()) {
					endTime = st.nextToken();
					values = storage.getValues(Long.parseLong(startTime), Long
							.parseLong(endTime));
					System.out.printf("getValues(%d,%d)", Long.parseLong(startTime), Long
							.parseLong(endTime));
				} else {
					values = storage.getValues(Long.parseLong(startTime));
					System.out.printf("getValues(%d)", Long.parseLong(startTime));
				}
				System.out.println("result size:" + values.size());

				
				
				out.print("!GET_VALUES_RESPONSE");

				for (Value value : values) {
					out.print(" (" + value.getTimestamp() + ","
							+ value.getValue() + ")");
				}
				
				System.out.println("step5");
				
				out.println();

			} else
				out.println("!ERROR: data store service not present!");
		} catch (Exception e) {
			out.println("!ERROR");
			logger.error(e.getMessage());
		}

	}

	private void getDirectoryReply(String msg) {

		if (store.getManager() != null) {
			List<String> directory = this.store.getManager()
					.getDeviceDirectory();

			out.print("!GET_DIRECTORY_RESPONSE");

			for (String label : directory) {
				out.print(" " + label);
			}

			out.println();
		} else
			out.println("!ERROR: data store service not present!");
	}

}
