package org.openmuc.mux.dataserver;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.Properties;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.openmuc.core.datamanager.Authentication;

/**
 * 
 * Simple ASCII based socket server for access to measured data.
 * The protocol is kept simple so it can be used my humans via telnet.
 * 
 * The IP and port the server will listen can be configured by the
 * OSGi/Java properties "org.openmuc.mux.dataserver.ip" and
 * "org.openmuc.mux.dataserver.port".
 * 
 * The default port is 9200. If no IP address is configuered the
 * server will listen to all IPs of the system.
 * 
 * @author mzillgit
 *
 */
public class DataSocketServer extends Thread {
	private static Logger logger = Logger.getLogger(DataSocketServer.class);
	
	public static int STATE_UNBOUND = 0;
	public static int STATE_LISTEN = 1;
	public static int STATE_ERROR = 2;
	
	private int serverState = STATE_UNBOUND;
	
	private ServerSocket serverSocket;
	private Properties props;
	
	private DataManagerServiceTracker store;
	private Vector<Connection> connectionList = new Vector<Connection>();
	
	/**
	 * 
	 * Create new Server and connect to a DataManagerServiceTracker
	 * 
	 * @param store
	 */
	public DataSocketServer(DataManagerServiceTracker store, Properties props) {
		this.store = store;
		this.props = props;
	}
	
	/**
	 * 
	 * Get the state of the server.
	 * 
	 * @return state one of STATE_UNBOUND, STATE_LISTEN or STATE_ERROR
	 */
	public int getServerState() {
		return serverState;
	}

	
	/**
	 * The main server thread.
	 */
	@Override
	public void run() {
		Authentication auth = new Authentication("default", "");
		
		try {
			String ipAddrStr = null;
			String portStr = null;
					
			ipAddrStr = props.getProperty("org.openmuc.mux.dataserver.ip");
			portStr = props.getProperty("org.openmuc.mux.dataserver.port");
			
			if (ipAddrStr == null) ipAddrStr = System.getProperty("org.openmuc.mux.dataserver.ip");
			if (portStr == null) portStr = System.getProperty("org.openmuc.mux.dataserver.port");
			
			logger.info("IP " + ipAddrStr);
			logger.info("Port " + portStr);
			
			int port;
			
			if (portStr != null) 
				port = new Integer(portStr).intValue();
			else port = 9200;
			
			
			if (ipAddrStr != null) {
				InetAddress inetAddr = InetAddress.getByName(ipAddrStr);
				
				serverSocket = new ServerSocket(port, 0, inetAddr);
			}
			else {
				serverSocket = new ServerSocket(port);
			}
			
			Socket clientSock;
			
			logger.info("DataSocketServer running...");
			
			try {
			
			while ((clientSock = serverSocket.accept()) != null) {
				Connection con = 
					new Connection(this.store, auth, clientSock, this);
				connectionList.add(con);
				con.start();
			}
			}
			catch (SocketException e) {
				
			}
			logger.info("DataSocketServer stopped.");
			
		} catch (IOException e) {
			e.printStackTrace();
			serverState = STATE_ERROR;
		}
		
		logger.info("DataSocketServer thread terminated!");
	}

	/** 
	 * 
	 * Callback method to be called by Connection objects if the connection
	 * is closed by timeout or on clients request.
	 * 
	 * @param con The calling Connection object
	 */
	protected void connectionClosed(Connection con) {
		connectionList.removeElement(con);
	}
	
	/**
	 *  Stop the server. Close all client sockets. Called by bundle activator.
	 */
	protected void stopServer() {
		logger.info("Close all open sockets");
		
		for (Connection con : connectionList) {
			con.close();
		}
		
		logger.info("Stop server.");
		try {
			serverSocket.close();
		} catch (IOException e) {
			logger.error(e.getMessage());
		}
		
		// this.interrupt();
	}
	
}
